from jax.ops import index, index_add
import jax.numpy as jnp
from math import pi
import numpy as np
from .curve import JaxCurve


def jaxHelicalfouriercurve_pure(dofs, quadpoints, order, n0, l0, R0, r0):
    A = dofs[:order]
    B = dofs[order:]
    phi = quadpoints*2*pi*l0
    m, phiV = jnp.meshgrid(jnp.arange(order), phi)
    AcosArray = jnp.sum(A*jnp.cos(m*phiV*n0/l0), axis=1)
    BsinArray = jnp.sum(B*jnp.sin(m*phiV*n0/l0), axis=1)
    eta = n0*phi/l0+AcosArray+BsinArray
    gamma = jnp.zeros((len(quadpoints), 3))
    gamma = index_add(gamma, index[:, 0], (R0+r0*jnp.cos(eta))*jnp.cos(phi))
    gamma = index_add(gamma, index[:, 1], (R0+r0*jnp.cos(eta))*jnp.sin(phi))
    gamma = index_add(gamma, index[:, 2], -r0*jnp.sin(eta))
    return gamma


class CurveHelical(JaxCurve):
    r'''Curve representation of a helical coil.
    The helical coil positions are specified by a poloidal angle eta that is a function of the toroidal angle phi with Fourier coefficients :math:`A_k` and :math:`B_k`.
    The poloidal angle is represented as 

    .. math:: 
        \eta = m_0 \phi/l_0 + \sum_k A_k \cos(n_0 \phi k/l_0) + B_k \sin(n_0 \phi k/l_0)

    Args:
        quadpoints: number of grid points/resolution along the curve;
        order:  number of fourier coefficients, i.e., the length of the array A (or B);
        n0:  toroidal periodicity/number of field periods
        l0:  number of :math:`2\pi` turns in :math:`\phi`
        R0:  major radius
        r0:  minor radius
    '''

    def __init__(self, quadpoints, order, n0, l0, R0, r0):
        if isinstance(quadpoints, int):
            quadpoints = np.linspace(0, 1, quadpoints, endpoint=False)
        pure = lambda dofs, points: jaxHelicalfouriercurve_pure(dofs, points, order, n0, l0, R0, r0)
        self.order = order
        self.n0 = n0
        self.l0 = l0
        self.R0 = R0
        self.r0 = r0
        self.coefficients = [np.zeros((order,)), np.zeros((order,))]
        super().__init__(quadpoints, pure)

    def num_dofs(self):
        return 2*self.order

    def get_dofs(self):
        return np.concatenate(self.coefficients)

    def set_dofs_impl(self, dofs):
        order = int(len(dofs)/2)
        for i in range(2):
            self.coefficients[i] = dofs[i*order:(i+1)*order]



import numpy as np
import simsoptpp as sopp

from .._core.optimizable import Optimizable


class Surface(Optimizable):
    r"""
    ``Surface`` is a base class for various representations of toroidal
    surfaces in simsopt.

    A ``Surface`` is modelled as a function :math:`\Gamma:[0, 1] \times [0, 1] \to R^3` and is evaluated at quadrature points :math:`\{\phi_1, \ldots, \phi_{n_\phi}\}\times\{\theta_1, \ldots, \theta_{n_\theta}\}`.

    """

    def __init__(self):
        Optimizable.__init__(self)
        self.dependencies = []
        self.fixed = np.full(len(self.get_dofs()), False)

    def plot(self, ax=None, show=True, plot_normal=False, plot_derivative=False, scalars=None, wireframe=True):
        """
        Plot the surface using mayavi. 
        Note: the `ax` and `show` parameter can be used to plot more than one surface:

        .. code-block::

            ax = surface1.plot(show=False)
            ax = surface2.plot(ax=ax, show=False)
            surface3.plot(ax=ax, show=True)


        """
        gamma = self.gamma()

        from mayavi import mlab
        mlab.mesh(gamma[:, :, 0], gamma[:, :, 1], gamma[:, :, 2], scalars=scalars)
        if wireframe:
            mlab.mesh(gamma[:, :, 0], gamma[:, :, 1], gamma[:, :, 2], representation='wireframe', color=(0, 0, 0), opacity=0.5)

        if plot_derivative:
            dg1 = 0.05 * self.gammadash1()
            dg2 = 0.05 * self.gammadash2()
            mlab.quiver3d(gamma[:, :, 0], gamma[:, :, 1], gamma[:, :, 2], dg1[:, :, 0], dg1[:, :, 1], dg1[:, :, 2])
            mlab.quiver3d(gamma[:, :, 0], gamma[:, :, 1], gamma[:, :, 2], dg2[:, :, 0], dg2[:, :, 1], dg2[:, :, 2])
        if plot_normal:
            n = 0.005 * self.normal()
            mlab.quiver3d(gamma[:, :, 0], gamma[:, :, 1], gamma[:, :, 2], n[:, :, 0], n[:, :, 1], n[:, :, 2])
        if show:
            mlab.show()

    def to_vtk(self, filename):
        from pyevtk.hl import gridToVTK
        g = self.gamma()
        ntor = g.shape[0]
        npol = g.shape[1]
        x = self.gamma()[:, :, 0].reshape((1, ntor, npol)).copy()
        y = self.gamma()[:, :, 1].reshape((1, ntor, npol)).copy()
        z = self.gamma()[:, :, 2].reshape((1, ntor, npol)).copy()
        n = self.normal().reshape((1, ntor, npol, 3))
        dphi = self.gammadash1().reshape((1, ntor, npol, 3))
        dtheta = self.gammadash2().reshape((1, ntor, npol, 3))
        contig = np.ascontiguousarray
        gridToVTK(filename, x, y, z, pointData={
            "dphi x dtheta": (contig(n[..., 0]), contig(n[..., 1]), contig(n[..., 2])),
            "dphi": (contig(dphi[..., 0]), contig(dphi[..., 1]), contig(dphi[..., 2])),
            "dtheta": (contig(dtheta[..., 0]), contig(dtheta[..., 1]), contig(dtheta[..., 2])),
        })

    def __repr__(self):
        return "Surface " + str(hex(id(self)))

    def to_RZFourier(self):
        """
        Return a :obj:`simsopt.geo.surfacerzfourier.SurfaceRZFourier` instance corresponding to the shape of this
        surface.  All subclasses should implement this abstract
        method.
        """
        raise NotImplementedError

    def cross_section(self, phi, thetas=None):
        """
        This function takes in a cylindrical angle :math:`\phi` and returns the cross
        section of the surface in that plane evaluated at `thetas`. This is
        done using the method of bisection.

        This function assumes that the surface intersection with the plane is a
        single curve.
        """

        # phi is assumed to be between [-pi, pi], so if it does not lie on that interval
        # we shift it by multiples of 2pi until it does
        phi = phi - np.sign(phi) * np.floor(np.abs(phi) / (2*np.pi)) * (2. * np.pi)
        if phi > np.pi:
            phi = phi - 2. * np.pi
        if phi < -np.pi:
            phi = phi + 2. * np.pi

        # varphi are the search intervals on which we look for the cross section in 
        # at constant cylindrical phi
        # The cross section is sampled at a number of points (theta_resolution) poloidally.
        varphi = np.asarray([0., 0.5, 1.0])

        if thetas is None:
            theta = np.asarray(self.quadpoints_theta)
        elif isinstance(thetas, np.ndarray):
            theta = thetas
        elif isinstance(thetas, int):
            theta = np.linspace(0, 1, thetas, endpoint=False)
        else:
            raise NotImplementedError('Need to pass int or 1d np.array to thetas')

        varphigrid, thetagrid = np.meshgrid(varphi, theta)
        varphigrid = varphigrid.T
        thetagrid = thetagrid.T

        # sample the surface at the varphi and theta points
        gamma = np.zeros((varphigrid.shape[0], varphigrid.shape[1], 3))
        self.gamma_lin(gamma, varphigrid.flatten(), thetagrid.flatten())

        # compute the cylindrical phi coordinate of each sampled point on the surface
        cyl_phi = np.arctan2(gamma[:, :, 1], gamma[:, :, 0])

        # reorder varphi, theta with respect to increasing cylindrical phi
        idx = np.argsort(cyl_phi, axis=0)
        cyl_phi = np.take_along_axis(cyl_phi, idx, axis=0)
        varphigrid = np.take_along_axis(varphigrid, idx, axis=0)

        # In case the target cylindrical angle "phi" lies above the first row or below the last row,
        # we must concatenate the lower row above the top row and the top row below the lower row.
        # This is allowable since the data in the matrices are periodic
        cyl_phi = np.concatenate((cyl_phi[-1, :][None, :]-2.*np.pi, cyl_phi, cyl_phi[0, :][None, :]+2.*np.pi), axis=0)
        varphigrid = np.concatenate((varphigrid[-1, :][None, :]-1., varphigrid, varphigrid[0, :][None, :]+1.), axis=0)

        # ensure that varphi does not have massive jumps.
        diff = varphigrid[1:]-varphigrid[:-1]
        pinc = np.abs(diff+1) < np.abs(diff)
        minc = np.abs(diff-1) < np.abs(diff)
        inc = pinc.astype(int) - minc.astype(int)
        prefix_sum = np.cumsum(inc, axis=0)
        varphigrid[1:] = varphigrid[1:] + prefix_sum

        # find the subintervals in varphi on which the desired cross section lies.
        # if idx_right == 0, then the subinterval must be idx_left = 0 and idx_right = 1
        idx_right = np.argmax(phi <= cyl_phi, axis=0)
        idx_right = np.where(idx_right == 0, 1, idx_right)
        idx_left = idx_right-1 

        varphi_left = varphigrid[idx_left, np.arange(idx_left.size)]
        varphi_right = varphigrid[idx_right, np.arange(idx_right.size)]
        cyl_phi_left = cyl_phi[idx_left, np.arange(idx_left.size)]
        cyl_phi_right = cyl_phi[idx_right, np.arange(idx_right.size)]

        # this function converts varphi to cylindrical phi, ensuring that the returned angle
        # lies between left_bound and right_bound.
        def varphi2phi(varphi_in, left_bound, right_bound):
            gamma = np.zeros((varphi_in.size, 3))
            self.gamma_lin(gamma, varphi_in, theta)
            phi = np.arctan2(gamma[:, 1], gamma[:, 0])
            pinc = (phi < left_bound).astype(int) 
            minc = (phi > right_bound).astype(int)
            phi = phi + 2.*np.pi * (pinc - minc)
            return phi

        def bisection(phia, a, phic, c):
            err = 1.
            while err > 1e-13:
                b = (a + c)/2.
                phib = varphi2phi(b, phia, phic)

                flag = (phib - phi) * (phic - phi) > 0
                # if flag is true,  then root lies on interval [a,b)
                # if flag is false, then root lies on interval [b,c]
                phia = np.where(flag, phia, phib)
                phic = np.where(flag, phib, phic)
                a = np.where(flag, a, b)
                c = np.where(flag, b, c)
                err = np.max(np.abs(a-c))
            b = (a + c)/2.
            return b          
        # bisect cyl_phi to compute the cross section
        sol = bisection(cyl_phi_left, varphi_left, cyl_phi_right, varphi_right)
        cross_section = np.zeros((sol.size, 3))
        self.gamma_lin(cross_section, sol, theta) 
        return cross_section

    def aspect_ratio(self):
        r"""
        Note: cylindrical coordinates are :math:`(R, \phi, Z)`, where :math:`\phi \in [-\pi,\pi)`
        and the angles that parametrize the surface are :math:`(\varphi, \theta) \in [0,1)^2`
        For a given surface, this function computes its aspect ratio using the VMEC
        definition:

        .. math::
            AR = R_{\text{major}} / R_{\text{minor}}

        where 

        .. math::
            R_{\text{minor}} &= \sqrt{ \overline{A} / \pi } \\
            R_{\text{major}} &= \frac{V}{2 \pi^2  R_{\text{minor}}^2} 

        and :math:`V` is the volume enclosed by the surface, and :math:`\overline{A}` is the
        average cross sectional area.
        The main difficult part of this calculation is the mean cross sectional
        area.  This is given by the integral

        .. math::
            \overline{A} = \frac{1}{2\pi} \int_{S_{\phi}} ~dS ~d\phi

        where :math:`S_\phi` is the cross section of the surface at the cylindrical angle :math:`\phi`.
        Note that :math:`\int_{S_\phi} ~dS` can be rewritten as a line integral 

        .. math::
            \int_{S_\phi}~dS &= \int_{S_\phi} ~dR dZ \\ 
            &= \int_{\partial S_\phi}  [R,0] \cdot \mathbf n/\|\mathbf n\| ~dl \\ 
            &= \int^1_{0} R \frac{\partial Z}{\partial \theta}~d\theta

        where :math:`\mathbf n = [n_R, n_Z] = [\partial Z/\partial \theta, -\partial R/\partial \theta]` is the outward pointing normal.

        Consider the surface in cylindrical coordinates terms of its angles :math:`[R(\varphi,\theta), 
        \phi(\varphi,\theta), Z(\varphi,\theta)]`.  The boundary of the cross section 
        :math:`\partial S_\phi` is given by the points :math:`\theta\rightarrow[R(\varphi(\phi,\theta),\theta),\phi, 
        Z(\varphi(\phi,\theta),\theta)]` for fixed :math:`\phi`.  The cross sectional area of :math:`S_\phi` becomes

        .. math::
            \int^{1}_{0} R(\varphi(\phi,\theta),\theta)
            \frac{\partial}{\partial \theta}[Z(\varphi(\phi,\theta),\theta)] ~d\theta

        Now, substituting this into the formula for the mean cross sectional area, we have

        .. math::
            \overline{A} = \frac{1}{2\pi}\int^{\pi}_{-\pi}\int^{1}_{0} R(\varphi(\phi,\theta),\theta)
                \frac{\partial}{\partial \theta}[Z(\varphi(\phi,\theta),\theta)] ~d\theta ~d\phi

        Instead of integrating over cylindrical :math:`\phi`, let's complete the change of variables and
        integrate over :math:`\varphi` using the mapping:

        .. math::
            [\phi,\theta] \leftarrow [\text{atan2}(y(\varphi,\theta), x(\varphi,\theta)), \theta]

        After the change of variables, the integral becomes:

        .. math::
            \overline{A} = \frac{1}{2\pi}\int^{1}_{0}\int^{1}_{0} R(\varphi,\theta) \left[\frac{\partial Z}{\partial \varphi} 
            \frac{\partial \varphi}{d \theta} + \frac{\partial Z}{\partial \theta} \right] \text{det} J ~d\theta ~d\varphi

        where :math:`\text{det}J` is the determinant of the mapping's Jacobian.

        """

        xyz = self.gamma()
        x2y2 = xyz[:, :, 0]**2 + xyz[:, :, 1]**2
        dgamma1 = self.gammadash1()
        dgamma2 = self.gammadash2()

        # compute the average cross sectional area
        J = np.zeros((xyz.shape[0], xyz.shape[1], 2, 2))
        J[:, :, 0, 0] = (xyz[:, :, 0] * dgamma1[:, :, 1] - xyz[:, :, 1] * dgamma1[:, :, 0])/x2y2
        J[:, :, 0, 1] = (xyz[:, :, 0] * dgamma2[:, :, 1] - xyz[:, :, 1] * dgamma2[:, :, 0])/x2y2
        J[:, :, 1, 0] = 0.
        J[:, :, 1, 1] = 1.

        detJ = np.linalg.det(J)
        Jinv = np.linalg.inv(J)

        dZ_dtheta = dgamma1[:, :, 2] * Jinv[:, :, 0, 1] + dgamma2[:, :, 2] * Jinv[:, :, 1, 1]
        mean_cross_sectional_area = np.abs(np.mean(np.sqrt(x2y2) * dZ_dtheta * detJ))/(2 * np.pi) 

        R_minor = np.sqrt(mean_cross_sectional_area / np.pi)
        R_major = np.abs(self.volume()) / (2. * np.pi**2 * R_minor**2)

        AR = R_major/R_minor
        return AR


def signed_distance_from_surface(xyz, surface):
    """
    Compute the signed distances from points ``xyz`` to a surface.  The sign is
    positive for points inside the volume surrounded by the surface.
    """
    gammas = surface.gamma().reshape((-1, 3))
    from scipy.spatial.distance import cdist
    dists = cdist(xyz, gammas)
    mins = np.argmin(dists, axis=1)
    n = surface.unitnormal().reshape((-1, 3))
    nmins = n[mins]
    gammamins = gammas[mins]

    # Now that we have found the closest node, we approximate the surface with
    # a plane through that node with the appropriate normal and then compute
    # the distance from the point to that plane
    # https://stackoverflow.com/questions/55189333/how-to-get-distance-from-point-to-plane-in-3d
    mindist = np.sum((xyz-gammamins) * nmins, axis=1)

    a_point_in_the_surface = np.mean(surface.gamma()[0, :, :], axis=0)
    sign_of_interiorpoint = np.sign(np.sum((a_point_in_the_surface-gammas[0, :])*n[0, :]))

    signed_dists = mindist * sign_of_interiorpoint
    return signed_dists


class SurfaceClassifier():
    r"""
    Takes in a toroidal surface and constructs an interpolant of the signed distance function
    :math:`f:R^3\to R` that is positive inside the volume contained by the surface,
    (approximately) zero on the surface, and negative outisde the volume contained by the surface.
    """

    def __init__(self, surface, p=1, h=0.05):
        """
        Args:
            surface: the surface to contruct the distance from.
            p: degree of the interpolant
            h: grid resolution of the interpolant
        """
        gammas = surface.gamma()
        r = np.linalg.norm(gammas[:, :, :2], axis=2)
        z = gammas[:, :, 2]
        rmin = max(np.min(r) - 0.1, 0.)
        rmax = np.max(r) + 0.1
        zmin = np.min(z) - 0.1
        zmax = np.max(z) + 0.1

        self.zrange = (zmin, zmax)
        self.rrange = (rmin, rmax)

        nr = int((self.rrange[1]-self.rrange[0])/h)
        nphi = int(2*np.pi/h)
        nz = int((self.zrange[1]-self.zrange[0])/h)

        def fbatch(rs, phis, zs):
            xyz = np.zeros((len(rs), 3))
            xyz[:, 0] = rs * np.cos(phis)
            xyz[:, 1] = rs * np.sin(phis)
            xyz[:, 2] = zs
            return list(signed_distance_from_surface(xyz, surface))

        rule = sopp.UniformInterpolationRule(p)
        self.dist = sopp.RegularGridInterpolant3D(
            rule, [rmin, rmax, nr], [0., 2*np.pi, nphi], [zmin, zmax, nz], 1, True)
        self.dist.interpolate_batch(fbatch)

    def evaluate(self, xyz):
        rphiz = np.zeros_like(xyz)
        rphiz[:, 0] = np.linalg.norm(xyz[:, :2], axis=1)
        rphiz[:, 1] = np.mod(np.arctan2(xyz[:, 1], xyz[:, 0]), 2*np.pi)
        rphiz[:, 2] = xyz[:, 2]
        d = np.zeros((xyz.shape[0], 1))
        self.dist.evaluate_batch(rphiz, d)
        return d

    def to_vtk(self, filename, h=0.01):
        from pyevtk.hl import gridToVTK

        nr = int((self.rrange[1]-self.rrange[0])/h)
        nphi = int(2*np.pi/h)
        nz = int((self.zrange[1]-self.zrange[0])/h)
        rs = np.linspace(self.rrange[0], self.rrange[1], nr)
        phis = np.linspace(0, 2*np.pi, nphi)
        zs = np.linspace(self.zrange[0], self.zrange[1], nz)

        R, Phi, Z = np.meshgrid(rs, phis, zs)
        X = R * np.cos(Phi)
        Y = R * np.sin(Phi)
        Z = Z

        RPhiZ = np.zeros((R.size, 3))
        RPhiZ[:, 0] = R.flatten()
        RPhiZ[:, 1] = Phi.flatten()
        RPhiZ[:, 2] = Z.flatten()
        vals = np.zeros((R.size, 1))
        self.dist.evaluate_batch(RPhiZ, vals)
        vals = vals.reshape(R.shape)
        gridToVTK(filename, X, Y, Z, pointData={"levelset": vals})



import numpy as np

import simsoptpp as sopp
from .curve import Curve


class CurveRZFourier(sopp.CurveRZFourier, Curve):
    r"""
    CurveRZFourier is a curve that is represented in cylindrical
       coordinates using the following Fourier series:

        .. math::
           r(\phi) &= \sum_{m=0}^{\text{order}} x_{c,m}\cos(n_{\text{fp}} m \phi) + \sum_{m=1}^{\text{order}} x_{s,m}\sin(n_{\text{fp}} m \phi) \\
           z(\phi) &= \sum_{m=0}^{\text{order}} z_{c,m}\cos(n_{\text{fp}} m \phi) + \sum_{m=1}^{\text{order}} z_{s,m}\sin(n_{\text{fp}} m \phi)

       If ``stellsym = True``, then the :math:`\sin` terms for :math:`r` and the :math:`\cos` terms for :math:`z` are zero.

       For the ``stellsym = False`` case, the dofs are stored in the order

        .. math::
           [r_{c,0}, \cdots, r_{c,\text{order}}, r_{s,1}, \cdots, r_{s,\text{order}}, z_{c,0},....]

       or in the ``stellsym = True`` case they are stored

        .. math::
           [r_{c,0},...,r_{c,order},z_{s,1},...,z_{s,order}]
    """

    def __init__(self, quadpoints, order, nfp, stellsym):
        if isinstance(quadpoints, int):
            quadpoints = list(np.linspace(0, 1./nfp, quadpoints, endpoint=False))
        elif isinstance(quadpoints, np.ndarray):
            quadpoints = list(quadpoints)
        sopp.CurveRZFourier.__init__(self, quadpoints, order, nfp, stellsym)
        Curve.__init__(self)

    def get_dofs(self):
        """
        This function returns the dofs associated to this object.
        """
        return np.asarray(sopp.CurveRZFourier.get_dofs(self))

    def set_dofs(self, dofs):
        """
        This function sets the dofs associated to this object.
        """
        sopp.CurveRZFourier.set_dofs(self, dofs)
        for d in self.dependencies:
            d.invalidate_cache()



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module provides several classes for representing toroidal
surfaces.  There is a base class Surface, and several child classes
corresponding to different discrete representations.
"""
import numpy as np
import logging
import abc
from jax.config import config
config.update("jax_enable_x64", True)  # Use double precision:

import jax.numpy as jnp
from jax import jacrev
#import xarray as xr

from .._core.util import isbool
from .._core.graph_optimizable import Optimizable

logger = logging.getLogger(__name__)


# @jit(static_argnums=(4, 5, 6, 7, 8, 9))
def area_volume_pure(rc, rs, zc, zs, stellsym, nfp, mpol, ntor, ntheta, nphi):
    """
    Compute the area and volume of a surface. This pure function is
    designed for automatic differentiation.
    """
    mdim = mpol + 1
    ndim = 2 * ntor + 1
    theta1d = jnp.linspace(0, 2 * jnp.pi, ntheta, endpoint=False)
    phi1d = jnp.linspace(0, 2 * jnp.pi / nfp, nphi, endpoint=False)
    dtheta = theta1d[1] - theta1d[0]
    dphi = phi1d[1] - phi1d[0]
    phi, theta = jnp.meshgrid(phi1d, theta1d)

    r = jnp.zeros((ntheta, nphi))
    x = jnp.zeros((ntheta, nphi))
    y = jnp.zeros((ntheta, nphi))
    z = jnp.zeros((ntheta, nphi))
    dxdtheta = jnp.zeros((ntheta, nphi))
    dydtheta = jnp.zeros((ntheta, nphi))
    dzdtheta = jnp.zeros((ntheta, nphi))
    dxdphi = jnp.zeros((ntheta, nphi))
    dydphi = jnp.zeros((ntheta, nphi))
    dzdphi = jnp.zeros((ntheta, nphi))
    sinphi = jnp.sin(phi)
    cosphi = jnp.cos(phi)
    for m in range(mdim):
        for jn in range(ndim):
            # Presently this loop includes negative n when m=0.
            # This is unnecesary but doesn't hurt I think.
            n_without_nfp = jn - ntor
            n = n_without_nfp * nfp
            angle = m * theta - n * phi
            sinangle = jnp.sin(angle)
            cosangle = jnp.cos(angle)
            rmnc = rc[m, jn]
            zmns = zs[m, jn]
            r += rmnc * cosangle
            x += rmnc * cosangle * cosphi
            y += rmnc * cosangle * sinphi
            z += zmns * sinangle

            dxdtheta += rmnc * (-m * sinangle) * cosphi
            dydtheta += rmnc * (-m * sinangle) * sinphi
            dzdtheta += zmns * m * cosangle

            dxdphi += rmnc * (n * sinangle * cosphi + cosangle * (-sinphi))
            dydphi += rmnc * (n * sinangle * sinphi + cosangle * cosphi)
            dzdphi += zmns * (-n * cosangle)
            if not stellsym:
                rmns = rs[m, jn]
                zmnc = zc[m, jn]
                r += rmns * sinangle
                x += rmns * sinangle * cosphi
                y += rmns * sinangle * sinphi
                z += zmnc * cosangle

                dxdtheta += rmns * (m * cosangle) * cosphi
                dydtheta += rmns * (m * cosangle) * sinphi
                dzdtheta += zmnc * (-m * sinangle)

                dxdphi += rmns * (-n * cosangle * cosphi + sinangle * (-sinphi))
                dydphi += rmns * (-n * cosangle * sinphi + sinangle * cosphi)
                dzdphi += zmnc * (n * sinangle)

    normalx = dydphi * dzdtheta - dzdphi * dydtheta
    normaly = dzdphi * dxdtheta - dxdphi * dzdtheta
    normalz = dxdphi * dydtheta - dydphi * dxdtheta
    norm_normal = jnp.sqrt(normalx * normalx + normaly * normaly + normalz * normalz)
    area = nfp * dtheta * dphi * jnp.sum(norm_normal)
    # Compute plasma volume using \int (1/2) R^2 dZ dphi
    # = \int (1/2) R^2 (dZ/dtheta) dtheta dphi
    volume = 0.5 * nfp * dtheta * dphi * jnp.sum(r * r * dzdtheta)
    return jnp.array([area, volume])


def dataset_area_volume(surface_rz_fourier,
                        stellsym,
                        nfp,
                        mpol,
                        ntor,
                        ntheta,
                        nphi):
    """
    Compute the area and volume of a surface. This pure function is
    designed for automatic differentiation.
    """
    mdim = mpol + 1
    ndim = 2 * ntor + 1
    theta1d = jnp.linspace(0, 2 * jnp.pi, ntheta, endpoint=False)
    phi1d = jnp.linspace(0, 2 * jnp.pi / nfp, nphi, endpoint=False)
    dtheta = theta1d[1] - theta1d[0]
    dphi = phi1d[1] - phi1d[0]
    phi, theta = jnp.meshgrid(phi1d, theta1d)

    r = jnp.zeros((ntheta, nphi))
    x = jnp.zeros((ntheta, nphi))
    y = jnp.zeros((ntheta, nphi))
    z = jnp.zeros((ntheta, nphi))
    dxdtheta = jnp.zeros((ntheta, nphi))
    dydtheta = jnp.zeros((ntheta, nphi))
    dzdtheta = jnp.zeros((ntheta, nphi))
    dxdphi = jnp.zeros((ntheta, nphi))
    dydphi = jnp.zeros((ntheta, nphi))
    dzdphi = jnp.zeros((ntheta, nphi))
    sinphi = jnp.sin(phi)
    cosphi = jnp.cos(phi)

    ds = surface_rz_fourier  # For simple access
    rc = ds["rc"]
    zs = ds["zs"]
    rs = ds.get("rs", None)
    zc = ds.get("zc", None)
    for m in range(mdim):
        for jn in range(ndim):
            # Presently this loop includes negative n when m=0.
            # This is unnecesary but doesn't hurt I think.
            n_without_nfp = jn - ntor
            n = n_without_nfp * nfp
            angle = m * theta - n * phi
            sinangle = jnp.sin(angle)
            cosangle = jnp.cos(angle)
            rmnc = rc[m, jn]
            zmns = zs[m, jn]
            r += rmnc * cosangle
            x += rmnc * cosangle * cosphi
            y += rmnc * cosangle * sinphi
            z += zmns * sinangle

            dxdtheta += rmnc * (-m * sinangle) * cosphi
            dydtheta += rmnc * (-m * sinangle) * sinphi
            dzdtheta += zmns * m * cosangle

            dxdphi += rmnc * (n * sinangle * cosphi + cosangle * (-sinphi))
            dydphi += rmnc * (n * sinangle * sinphi + cosangle * cosphi)
            dzdphi += zmns * (-n * cosangle)
            if not stellsym:
                rmns = rs[m, jn]
                zmnc = zc[m, jn]
                r += rmns * sinangle
                x += rmns * sinangle * cosphi
                y += rmns * sinangle * sinphi
                z += zmnc * cosangle

                dxdtheta += rmns * (m * cosangle) * cosphi
                dydtheta += rmns * (m * cosangle) * sinphi
                dzdtheta += zmnc * (-m * sinangle)

                dxdphi += rmns * (-n * cosangle * cosphi + sinangle * (-sinphi))
                dydphi += rmns * (-n * cosangle * sinphi + sinangle * cosphi)
                dzdphi += zmnc * (n * sinangle)

    normalx = dydphi * dzdtheta - dzdphi * dydtheta
    normaly = dzdphi * dxdtheta - dxdphi * dzdtheta
    normalz = dxdphi * dydtheta - dydphi * dxdtheta
    norm_normal = jnp.sqrt(normalx * normalx + normaly * normaly + normalz * normalz)
    area = nfp * dtheta * dphi * jnp.sum(norm_normal)
    # Compute plasma volume using \int (1/2) R^2 dZ dphi
    # = \int (1/2) R^2 (dZ/dtheta) dtheta dphi
    volume = 0.5 * nfp * dtheta * dphi * jnp.sum(r * r * dzdtheta)
    return jnp.array([area, volume])


# area_volume_pure(rc, rs, zc, zs, stellsym, nfp, mpol, ntor, ntheta, nphi)
# jit_area_volume_pure = jit(area_volume_pure, static_argnums=(4, 5, 6, 7, 8, 9))
# jit_area_volume_pure = jit(area_volume_pure, static_argnums=(8, 9))
jit_area_volume_pure = area_volume_pure
darea_volume_pure = jacrev(area_volume_pure, argnums=(0, 1, 2, 3))

jit_dataset_area_volume = dataset_area_volume


class Surface(abc.ABC):
    """
    Surface is a base class for various representations of toroidal
    surfaces in simsopt.
    """

    def __init__(self, nfp=1, stellsym=True):
        #if not isinstance(nfp, int):
        #    raise TypeError('nfp must be an integer')
        #if not isbool(stellsym):
        #    raise TypeError('stellsym must be a bool')
        self.nfp = nfp
        self.stellsym = stellsym

    def __repr__(self):
        return "Surface " + str(hex(id(self))) + " (nfp=" + str(self.nfp) \
               + ", stellsym=" + str(self.stellsym) + ")"

    @abc.abstractmethod
    def to_RZFourier(self):
        """
        Return a SurfaceRZFourier instance corresponding to the shape of this
        surface.  All subclasses should implement this abstract
        method.
        """
        #raise NotImplementedError


class SurfaceRZFourier(Surface, Optimizable):
    """
    SurfaceRZFourier is a surface that is represented in cylindrical
    coordinates using the following Fourier series: 

    r(theta, phi) = \sum_{m=0}^{mpol} \sum_{n=-ntor}^{ntor} [
                     r_{c,m,n} \cos(m \theta - n nfp \phi)
                     + r_{s,m,n} \sin(m \theta - n nfp \phi) ]

    and the same for z(theta, phi).

    Here, (r, phi, z) are standard cylindrical coordinates, and theta
    is any poloidal angle.
    """

    def __init__(self, nfp=1, stellsym=True, mpol=1, ntor=0):

        Surface.__init__(self, nfp=nfp, stellsym=stellsym)
        if mpol < 1:
            raise ValueError("mpol must be at least 1")
        if ntor < 0:
            raise ValueError("ntor must be at least 0")
        self.mpol = mpol
        self.ntor = ntor

        logger.info("Allocating SurfaceRZFourier")
        self.mdim = self.mpol + 1
        self.ndim = 2 * self.ntor + 1
        self.size = self.mdim * self.ndim
        self.shape = (self.mdim, self.ndim)

        rc = np.zeros(self.shape)
        zs = np.zeros(self.shape)
        dof_names = self.make_names('rc') + self.make_names('zs')

        # Initialize to an axisymmetric torus with major radius 1m and
        # minor radius 0.1m
        rc[0, ntor] = 1.0
        rc[1, ntor] = 0.1
        zs[1, ntor] = 0.1

        dofs_x = np.concatenate([rc.flatten(), zs.flatten()])
        rc_fixed = [True] * self.ntor + [False] * (self.ntor + 1)
        zs_fixed = [True] * (self.ntor + 1) + [False] * self.ntor
        for m in range(self.mpol):
            rc_fixed += [False] * (2*self.ntor + 1)
            zs_fixed += [False] * (2*self.ntor + 1)
        dofs_fixed = np.concatenate([np.array(rc_fixed), np.array(zs_fixed)])

        if not self.stellsym:
            rs = np.zeros(self.shape)
            zc = np.zeros(self.shape)
            dofs_x = np.concatenate([dofs_x, rs.flatten(), zc.flatten()])

            zc_fixed = [True] * self.ntor + [False] * (self.ntor + 1)
            rs_fixed = [True] * (self.ntor + 1) + [False] * self.ntor
            for m in range(self.mpol):
                zc_fixed += [False] * (2 * self.ntor + 1)
                rs_fixed += [False] * (2 * self.ntor + 1)
            dofs_fixed = np.concatenate([dofs_fixed, np.array(rs_fixed),
                                        np.array(zc_fixed)])
            dof_names += self.make_names('rs') + self.make_names('zc')

        Optimizable.__init__(self, x0=dofs_x, names=dof_names, fixed=dofs_fixed)

        #self.recalculate = True
        self.recalculate_derivs = True

        # Resolution for computing area, volume, etc:
        self.ntheta = 63
        self.nphi = 62

    def make_names(self, prefix):
        """
        Create the arrays for the rc, rs, zc, and zs coefficients.
        """
        names = []
        for m in range(self.mpol + 1):
            names += [prefix + '(' + str(m) + ',' + str(n) + ')' \
                      for n in range(-self.ntor, self.ntor + 1)]
        return names

    def __repr__(self):
        return "SurfaceRZFourier " + str(hex(id(self))) + " (nfp=" + \
               str(self.nfp) + ", stellsym=" + str(self.stellsym) + \
               ", mpol=" + str(self.mpol) + ", ntor=" + str(self.ntor) \
               + ")"

    def _validate_mn(self, m, n):
        """
        Check whether m and n are in the allowed range.
        """
        if m < 0:
            raise ValueError('m must be >= 0')
        if m > self.mpol:
            raise ValueError('m must be <= mpol')
        if n > self.ntor:
            raise ValueError('n must be <= ntor')
        if n < -self.ntor:
            raise ValueError('n must be >= -ntor')

    @property
    def rc(self):
        return self.local_full_x[0:self.size].reshape(self.shape)

    @rc.setter
    def rc(self, rc):
        rc = np.array(rc)
        self._dofs.iloc[0:self.size, 0] = rc.flatten()
        self.new_x = True

    @property
    def zs(self):
        return self.local_full_x[self.size:2*self.size].reshape(self.shape)

    @zs.setter
    def zs(self, zs):
        zs = np.array(zs)
        self._dofs.iloc[self.size:2*self.size, 0] = zs.flatten()
        self.new_x = True

    @property
    def rs(self):
        if self.stellsym:
            return None
        return self.local_full_x[2*self.size:3*self.size].reshape(self.shape)

    @rs.setter
    def rs(self, rs):
        rs = np.array(rs)
        self._dofs.iloc[2*self.size:3*self.size, 0] = rs.flatten()
        self.new_x = True

    @property
    def zc(self):
        if self.stellsym:
            return None
        return self.local_full_x[3*self.size:].reshape(self.shape)

    @zc.setter
    def zc(self, zc):
        zc = np.array(zc)
        self._dofs.iloc[3*self.size:, 0] = zc.flatten()
        self.new_x = True

    def get_rc(self, m, n):
        """
        Return a particular rc Parameter.
        """
        self._validate_mn(m, n)
        rc = self.local_full_x[0:self.size].reshape(self.shape)
        return rc[m, n + self.ntor]

    def get_rs(self, m, n):
        """
        Return a particular rs Parameter.
        """
        if self.stellsym:
            return ValueError(
                'rs does not exist for this stellarator-symmetric surface.')
        self._validate_mn(m, n)
        rs = self.local_full_x[2*self.size:3*self.size].reshape(self.shape)
        return rs[m, n + self.ntor]

    def get_zc(self, m, n):
        """
        Return a particular zc Parameter.
        """
        if self.stellsym:
            return ValueError(
                'zc does not exist for this stellarator-symmetric surface.')
        self._validate_mn(m, n)
        zc = self.local_full_x[3*self.size:].reshape(self.shape)
        return zc[m, n + self.ntor]

    def get_zs(self, m, n):
        """
        Return a particular zs Parameter.
        """
        self._validate_mn(m, n)
        zs = self.local_full_x[self.size:2*self.size].reshape(self.shape)
        return zs[m, n + self.ntor]

    def set_rc(self, m, n, val):
        """
        Set a particular rc Parameter.
        """
        self._validate_mn(m, n)
        i = m * self.ndim + n + self.ntor
        self._dofs.iloc[i, 0] = val
        #self.rc[m, n + self.ntor] = val
        #self.recalculate = True
        self.new_x = True
        self.recalculate_derivs = True

    def set_rs(self, m, n, val):
        """
        Set a particular rs Parameter.
        """
        if self.stellsym:
            return ValueError(
                'rs does not exist for this stellarator-symmetric surface.')
        self._validate_mn(m, n)
        i = m * self.ndim + n + self.ntor + 2 * self.size
        #self.rs[m, n + self.ntor] = val
        self._dofs.iloc[i, 0] = val
        #self.recalculate = True
        self.new_x = True
        self.recalculate_derivs = True

    def set_zc(self, m, n, val):
        """
        Set a particular zc Parameter.
        """
        if self.stellsym:
            return ValueError(
                'zc does not exist for this stellarator-symmetric surface.')
        self._validate_mn(m, n)
        #self.zc[m, n + self.ntor] = val
        i = m * self.ndim + n + self.ntor + 3 * self.size
        self._dofs.iloc[i, 0] = val
        #self.recalculate = True
        self.new_x = True
        self.recalculate_derivs = True

    def set_zs(self, m, n, val):
        """
        Set a particular zs Parameter.
        """
        self._validate_mn(m, n)
        i = m * self.ndim + n + self.ntor + self.size
        self._dofs.iloc[i, 0] = val
        self.new_x = True
        self.recalculate_derivs = True

    def area_volume(self):
        """
        Compute the surface area and the volume enclosed by the surface.
        """
        if self.new_x:
            logger.info('Running calculation of area and volume')
        else:
            logger.info('area_volume called, but no need to recalculate')
            return

        rc = self.local_full_x[0:self.size].reshape(self.shape)
        zs = self.local_full_x[self.size:2*self.size].reshape(self.shape)

        if self.stellsym:
            rs = None
            zc = None
        else:
            #rs = self.rs
            #zc = self.zc
            rs = self.local_full_x[2*self.size:3*self.size].reshape(self.shape)
            zc = self.local_full_x[3*self.size:].reshape(self.shape)

        results = jit_area_volume_pure(rc, rs, zc, zs,
                                       self.stellsym, self.nfp, self.mpol,
                                       self.ntor, self.ntheta, self.nphi)

        self._area = float(results[0])
        self._volume = float(results[1])
        self.new_x = False
        return results

    def area(self):
        """
        Return the area of the surface.
        """
        self.area_volume()
        return self._area

    def f(self):
        return self.area_volume()

    def volume(self):
        """
        Return the volume of the surface.
        """
        self.area_volume()
        return self._volume

    return_fn_map = {'area': area, 'volume': volume}

    def darea_volume(self):
        """
        Compute the derivative of the surface area and the volume enclosed
        by the surface.
        """
        if self.recalculate_derivs:
            logger.info('Running calculation of derivative of area and volume')
        else:
            logger.info('darea_volume called, but no need to recalculate')
            return

        self.recalculate_derivs = False

        rc = self.local_full_x[0:self.size].reshape(self.shape)
        zs = self.local_full_x[self.size:2*self.size].reshape(self.shape)
        if self.stellsym:
            rs = None
            zc = None
        else:
            rs = self.local_full_x[2*self.size:3*self.size].reshape(self.shape)
            zc = self.local_full_x[3*self.size:].reshape(self.shape)

        results = darea_volume_pure(rc, rs, zc, zs,
                                    self.stellsym, self.nfp, self.mpol,
                                    self.ntor, self.ntheta, self.nphi)

        self._darea_drc = np.array(results[0][0, :, :])
        self._dvolume_drc = np.array(results[0][1, :, :])
        self._darea_dzs = np.array(results[3][0, :, :])
        self._dvolume_dzs = np.array(results[3][1, :, :])
        if not self.stellsym:
            self._darea_drs = np.array(results[1][0, :, :])
            self._dvolume_drs = np.array(results[1][1, :, :])
            self._darea_dzc = np.array(results[2][0, :, :])
            self._dvolume_dzc = np.array(results[2][1, :, :])

    def darea(self):
        """
        Return the area of the surface.
        """
        self.darea_volume()
        mpol = self.mpol  # Shorthand
        ntor = self.ntor
        if self.stellsym:
            return np.concatenate(
                (self._darea_drc[0, ntor:],
                 self._darea_drc[1:, :].reshape(mpol * (ntor * 2 + 1)),
                 self._darea_dzs[0, ntor + 1:],
                 self._darea_dzs[1:, :].reshape(mpol * (ntor * 2 + 1))))
        else:
            return np.concatenate(
                (self._darea_drc[0, ntor:],
                 self._darea_drc[1:, :].reshape(mpol * (ntor * 2 + 1)),
                 self._darea_dzs[0, ntor + 1:],
                 self._darea_dzs[1:, :].reshape(mpol * (ntor * 2 + 1)),
                 self._darea_drs[0, ntor + 1:],
                 self._darea_drs[1:, :].reshape(mpol * (ntor * 2 + 1)),
                 self._darea_dzc[0, ntor:],
                 self._darea_dzc[1:, :].reshape(mpol * (ntor * 2 + 1))))

    def dvolume(self):
        """
        Return the volume of the surface.
        """
        self.darea_volume()
        mpol = self.mpol  # Shorthand
        ntor = self.ntor
        if self.stellsym:
            return np.concatenate(
                (self._dvolume_drc[0, ntor:],
                 self._dvolume_drc[1:, :].reshape(mpol * (ntor * 2 + 1)),
                 self._dvolume_dzs[0, ntor + 1:],
                 self._dvolume_dzs[1:, :].reshape(mpol * (ntor * 2 + 1))))
        else:
            return np.concatenate(
                (self._dvolume_drc[0, ntor:],
                 self._dvolume_drc[1:, :].reshape(mpol * (ntor * 2 + 1)),
                 self._dvolume_dzs[0, ntor + 1:],
                 self._dvolume_dzs[1:, :].reshape(mpol * (ntor * 2 + 1)),
                 self._dvolume_drs[0, ntor + 1:],
                 self._dvolume_drs[1:, :].reshape(mpol * (ntor * 2 + 1)),
                 self._dvolume_dzc[0, ntor:],
                 self._dvolume_dzc[1:, :].reshape(mpol * (ntor * 2 + 1))))

    @classmethod
    def from_focus(cls, filename):
        """
        Read in a surface from a FOCUS-format file.
        """
        f = open(filename, 'r')
        lines = f.readlines()
        f.close()

        # Read the line containing Nfou and nfp:
        splitline = lines[1].split()
        errmsg = "This does not appear to be a FOCUS-format file."
        assert len(splitline) == 3, errmsg
        Nfou = int(splitline[0])
        nfp = int(splitline[1])

        # Now read the Fourier amplitudes:
        n = np.full(Nfou, 0)
        m = np.full(Nfou, 0)
        rc = np.zeros(Nfou)
        rs = np.zeros(Nfou)
        zc = np.zeros(Nfou)
        zs = np.zeros(Nfou)
        for j in range(Nfou):
            splitline = lines[j + 4].split()
            n[j] = int(splitline[0])
            m[j] = int(splitline[1])
            rc[j] = float(splitline[2])
            rs[j] = float(splitline[3])
            zc[j] = float(splitline[4])
            zs[j] = float(splitline[5])
        assert np.min(m) == 0
        stellsym = np.max(np.abs(rs)) == 0 and np.max(np.abs(zc)) == 0
        mpol = int(np.max(m))
        ntor = int(np.max(np.abs(n)))

        surf = cls(nfp=nfp, stellsym=stellsym, mpol=mpol, ntor=ntor)
        for j in range(Nfou):
            surf.set_rc(m[j], n[j], rc[j])
            surf.set_zs(m[j], n[j], zs[j])
            if not stellsym:
                surf.set_rs(m[j], n[j], rs[j])
                surf.set_zc(m[j], n[j], zc[j])

        return surf

    #def get_dofs(self):
    #    """
    #    Return a 1D numpy array with all the degrees of freedom.
    #    """
    #    mpol = self.mpol  # Shorthand
    #    ntor = self.ntor
    #    if self.stellsym:
    #        return np.concatenate(
    #            (self.rc[0, ntor:],
    #             self.rc[1:, :].reshape(mpol * (ntor * 2 + 1)),
    #             self.zs[0, ntor + 1:],
    #             self.zs[1:, :].reshape(mpol * (ntor * 2 + 1))))
    #    else:
    #        return np.concatenate(
    #            (self.rc[0, ntor:],
    #             self.rc[1:, :].reshape(mpol * (ntor * 2 + 1)),
    #             self.zs[0, ntor + 1:],
    #             self.zs[1:, :].reshape(mpol * (ntor * 2 + 1)),
    #             self.rs[0, ntor + 1:],
    #             self.rs[1:, :].reshape(mpol * (ntor * 2 + 1)),
    #             self.zc[0, ntor:],
    #             self.zc[1:, :].reshape(mpol * (ntor * 2 + 1))))

    #def set_dofs(self, v):
    #    """
    #    Set the shape coefficients from a 1D list/array
    #    """

    #    n = len(self.get_dofs())
    #    if len(v) != n:
    #        raise ValueError('Input vector should have ' + str(n) + \
    #                         ' elements but instead has ' + str(len(v)))

    #    # Check whether any elements actually change:
    #    if np.all(np.abs(self.get_dofs() - np.array(v)) == 0):
    #        logger.info('set_dofs called, but no dofs actually changed')
    #        return

    #    logger.info('set_dofs called, and at least one dof changed')
    #    self.recalculate = True
    #    self.recalculate_derivs = True

    #    mpol = self.mpol  # Shorthand
    #    ntor = self.ntor
    #    self.rc[0, ntor:] = v[0:ntor + 1]
    #    self.rc[1:, :] = np.array(v[ntor + 1:ntor + 1 + mpol * (ntor * 2 + 1)]).reshape(mpol, ntor * 2 + 1)
    #    self.zs[0, ntor + 1:] = v[ntor + 1 + mpol * (ntor * 2 + 1): \
    #                              ntor * 2 + 1 + mpol * (ntor * 2 + 1)]
    #    self.zs[1:, :] = np.array(v[ntor * 2 + 1 + mpol * (ntor * 2 + 1): \
    #                                       ntor * 2 + 1 + 2 * mpol * (ntor * 2 + 1)]).reshape(mpol, ntor * 2 + 1)

    def to_RZFourier(self):
        """
        No conversion necessary.
        """
        return self

    def to_Garabedian(self):
        """
        Return a SurfaceGarabedian object with the identical shape.

        For a derivation of the transformation here, see 
        https://terpconnect.umd.edu/~mattland/assets/notes/toroidal_surface_parameterizations.pdf
        """
        if not self.stellsym:
            raise RuntimeError('Non-stellarator-symmetric SurfaceGarabedian '
                               'objects have not been implemented')
        mmax = self.mpol + 1
        mmin = np.min((0, 1 - self.mpol))
        s = SurfaceGarabedian(nfp=self.nfp, mmin=mmin, mmax=mmax,
                              nmin=-self.ntor, nmax=self.ntor)
        for n in range(-self.ntor, self.ntor + 1):
            for m in range(mmin, mmax + 1):
                Delta = 0
                if m - 1 >= 0:
                    Delta = 0.5 * (self.get_rc(m - 1, n) - self.get_zs(m - 1, n))
                if 1 - m >= 0:
                    Delta += 0.5 * (self.get_rc(1 - m, -n) + self.get_zs(1 - m, -n))
                s.set_Delta(m, n, Delta)

        return s


class SurfaceGarabedian(Surface, Optimizable):
    """
    SurfaceGarabedian represents a toroidal surface for which the
    shape is parameterized using Garabedian's Delta_{m,n}
    coefficients.

    The present implementation assumes stellarator symmetry. Note that
    non-stellarator-symmetric surfaces require that the Delta_{m,n}
    coefficients be imaginary.
    """

    def __init__(self, nfp=1, mmax=1, mmin=0, nmax=0, nmin=None):
        if nmin is None:
            nmin = -nmax
        # Perform some validation.
        if mmax < mmin:
            raise ValueError("mmin must be >= mmax")
        if nmax < nmin:
            raise ValueError("nmin must be >= nmax")
        if mmax < 1:
            raise ValueError("mmax must be >= 1")
        if mmin > 0:
            raise ValueError("mmin must be <= 0")
        Surface.__init__(self, nfp=nfp, stellsym=True)
        self.mmin = mmin
        self.mmax = mmax
        self.nmin = nmin
        self.nmax = nmax
        #self.allocate()

        # Create the array for the Delta_{m,n} coefficients.
        logger.info("Allocating SurfaceGarabedian")
        self.mdim = self.mmax - self.mmin + 1
        self.ndim = self.nmax - self.nmin + 1
        self.shape = (self.mdim, self.ndim)

        Delta = np.zeros(self.shape)
        dof_names = []
        for m in range(self.mmin, self.mmax + 1):
            for n in range(self.nmin, self.nmax + 1):
                dof_names.append('Delta(' + str(m) + ',' + str(n) + ')')

        #self.new_x = True
        self.recalculate_derivs = True

        Optimizable.__init__(self, Delta.flatten(), names=dof_names)
        # Initialize to an axisymmetric torus with major radius 1m and
        # minor radius 0.1m
        self.set_Delta(1, 0, 1.0)
        self.set_Delta(0, 0, 0.1)

    def __repr__(self):
        return "SurfaceGarabedian " + str(hex(id(self))) + " (nfp=" + \
            str(self.nfp) + ", mmin=" + str(self.mmin) + ", mmax=" + str(self.mmax) \
            + ", nmin=" + str(self.nmin) + ", nmax=" + str(self.nmax) \
            + ")"

    @property
    def Delta(self):
        return self.local_full_x.reshape(self.shape)

    @Delta.setter
    def Delta(self, Delta):
        assert(self.shape == Delta.shape)
        self._dofs.loc[:, 'x'] = Delta.flatten()

    def get_Delta(self, m, n):
        """
        Return a particular Delta_{m,n} coefficient.
        """
        return self.Delta[m - self.mmin, n - self.nmin]

    def set_Delta(self, m, n, val):
        """
        Set a particular Delta_{m,n} coefficient.
        """
        # Not sure if the below statement will not lead to a copy under all
        # circumstances. TODO: Investigate this
        #self.Delta[m - self.mmin, n - self.nmin] = val

        # Using a safer approach
        i = self.ndim * (m - self.mmin) + n - self.nmin
        self._dofs.iloc[i, 0] = val
        self.new_x = True
        self.recalculate_derivs = True

    #def get_dofs(self):
    #    """
    #    Return a 1D numpy array with all the degrees of freedom.
    #    """
    #    num_dofs = (self.mmax - self.mmin + 1) * (self.nmax - self.nmin + 1)
    #    return np.reshape(self.Delta, (num_dofs,), order='F')

    #def set_dofs(self, v):
    #    """
    #    Set the shape coefficients from a 1D list/array
    #    """

    #    n = len(self.get_dofs())
    #    if len(v) != n:
    #        raise ValueError('Input vector should have ' + str(n) + \
    #                         ' elements but instead has ' + str(len(v)))

    #    # Check whether any elements actually change:
    #    if np.all(np.abs(self.get_dofs() - np.array(v)) == 0):
    #        logger.info('set_dofs called, but no dofs actually changed')
    #        return

    #    logger.info('set_dofs called, and at least one dof changed')
    #    self.recalculate = True
    #    self.recalculate_derivs = True

    #    self.Delta = v.reshape((self.mmax - self.mmin + 1, self.nmax - self.nmin + 1), order='F')

    def to_RZFourier(self):
        """
        Return a SurfaceRZFourier object with the identical shape.

        For a derivation of the transformation here, see 
        https://terpconnect.umd.edu/~mattland/assets/notes/toroidal_surface_parameterizations.pdf
        """
        mpol = int(np.max((1, self.mmax - 1, 1 - self.mmin)))
        ntor = int(np.max((self.nmax, -self.nmin)))
        s = SurfaceRZFourier(nfp=self.nfp, stellsym=True, mpol=mpol, ntor=ntor)
        s.set_rc(0, 0, self.get_Delta(1, 0))
        for m in range(mpol + 1):
            nmin = -ntor
            if m == 0:
                nmin = 1
            for n in range(nmin, ntor + 1):
                Delta1 = 0
                Delta2 = 0
                if 1 - m >= self.mmin and -n >= self.nmin and -n <= self.nmax:
                    Delta1 = self.get_Delta(1 - m, -n)
                if 1 + m <= self.mmax and n >= self.nmin and n <= self.nmax:
                    Delta2 = self.get_Delta(1 + m, n)
                s.set_rc(m, n, Delta1 + Delta2)
                s.set_zs(m, n, Delta1 - Delta2)

        return s

    def area_volume(self):
        """
        Compute the surface area and the volume enclosed by the surface.
        """
        if self.new_x:
            logger.info('Running calculation of area and volume')
        else:
            logger.info('area_volume called, but no need to recalculate')
            return

        self.new_x = False

        # Delegate to the area and volume calculations of SurfaceRZFourier():
        s = self.to_RZFourier()
        #self._area = s.area()
        #self._volume = s.volume()
        return s.area_volume()

    def f(self):
        return self.area_volume()

    def area(self):
        """
        Return the area of the surface.
        """
        return self.area_volume()[0]
        #return self._area

    def volume(self):
        """
        Return the volume of the surface.
        """
        return self.area_volume()[1]
        #return self._volume

    return_fn_map = {'area': area, 'volume': volume}




from benchmark_xtensor_python import rect_to_polar
import numpy as np

from timeit import timeit
w = np.ones(100000, dtype=complex)
print (timeit('rect_to_polar(w[::2])', 'from __main__ import w, rect_to_polar', number=1000))



PROTON_MASS = 1.67262192369e-27  # kg
NEUTRON_MASS = 1.67492749804e-27  # kg
ELEMENTARY_CHARGE = 1.602176634e-19  # C
ONE_EV = 1.602176634e-19  # J
ALPHA_PARTICLE_MASS = 2*PROTON_MASS + 2*NEUTRON_MASS
ALPHA_PARTICLE_CHARGE = 2*ELEMENTARY_CHARGE
FUSION_ALPHA_PARTICLE_ENERGY = 3.52e6 * ONE_EV



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
Provides graph based Optimizable class, whose instances can be used to
build an optimization problem in a graph like manner.
"""

from __future__ import annotations

import types
import hashlib
from collections.abc import Callable as ABC_Callable, Hashable
from collections import defaultdict
from numbers import Real, Integral
from typing import Union, Tuple, Dict, Callable, Sequence, \
    MutableSequence as MutSeq, List

import numpy as np
import pandas as pd
from deprecated import deprecated

from ..util.types import RealArray, StrArray, BoolArray, Key
from .util import ImmutableId, OptimizableMeta


class DOFs(pd.DataFrame):
    """
    Defines the (D)egrees (O)f (F)reedom(s) associated with optimization

    This class holds data related to the degrees of freedom
    associated with an Optimizable object. The class subclasses
    pandas.DataFrame. To access the data stored in the DOFs class as a
    pandas dataframe, use the labels under internal column shown in the
    table below.

    DOFs Dataframe column index table

    =====   =============  ===============
    Index   External name  Internal column
    =====   =============  ===============
    0       x              _x
    1       free           free
    2       lower_bounds   _lb
    3       upper_bounds   _ub
    =====   =============  ===============

    The class implements the external name column properties in the above
    table as properties. Additional methods to update bounds, fix/unfix DOFs,
    etc. are also defined.
    """

    def __init__(self,
                 x: RealArray = None,  # To enable empty DOFs object
                 names: StrArray = None,
                 free: BoolArray = None,
                 lower_bounds: RealArray = None,
                 upper_bounds: RealArray = None) -> None:
        """
        Args:
            x: Numeric values of the DOFs
            names: Names of the dofs
            free: Array of boolean values denoting if the DOFs is are free.
                  False values implies the corresponding DOFs are fixed
            lower_bounds: Lower bounds for the DOFs. Meaningful only if
                DOF is not fixed. Default is np.NINF
            upper_bounds: Upper bounds for the DOFs. Meaningful only if
                DOF is not fixed. Default is np.inf
        """
        if x is None:
            x = np.array([])
        else:
            x = np.array(x, dtype=np.double)
        if names is None:
            names = ["x{}".format(i) for i in range(len(x))]

        if free is not None:
            free = np.array(free, dtype=np.bool_)
        else:
            free = np.full(len(x), True)

        if lower_bounds is not None:
            lb = np.array(lower_bounds, np.double)
        else:
            lb = np.full(len(x), np.NINF)

        if upper_bounds is not None:
            ub = np.array(upper_bounds, np.double)
        else:
            ub = np.full(len(x), np.inf)
        super().__init__(data={"_x": x, "free": free, "_lb": lb, "_ub": ub},
                         index=names)

    def fix(self, key: Key) -> None:
        """
        Fixes the specified DOF

        Args:
            key: Key to identify the DOF
        """
        if isinstance(key, str):
            self.loc[key, 'free'] = False
        else:
            self.iloc[key, 1] = False

    def unfix(self, key: Key) -> None:
        """
        Unfixes the specified DOF

        Args:
            key: Key to identify the DOF
        """
        if isinstance(key, str):
            self.loc[key, 'free'] = True
        else:
            self.iloc[key, 1] = True

    def fix_all(self) -> None:
        """
        Fixes all the DOFs
        """
        self.free = self.free.apply(lambda x: False)

    def unfix_all(self) -> None:
        """
        Makes all DOFs variable
        Caution: Make sure the bounds are well defined
        """
        self.free = self.free.apply(lambda x: True)

    def any_free(self) -> bool:
        """
        Checks for any free DOFs

        Returns:
            True if any free DOF is found, else False
        """
        return self.free.any()

    def any_fixed(self) -> bool:
        """
        Checks for any free DOFs

        Returns:
            True if any fixed DOF is found, else False
        """
        return not self.free.all()

    def all_free(self) -> bool:
        """
        Checks if all DOFs are allowed to be varied

        Returns:
            True if all DOFs are free to changed
        """
        return self.free.all()

    def all_fixed(self) -> bool:
        """
        Checks if all the DOFs are fixed

        Returns:
            True if all DOFs are fixed
        """
        return not self.free.any()

    @property
    def x(self) -> RealArray:
        """

        Returns:
            The values of the free DOFs.
        """
        return self.loc[self.free, "_x"].to_numpy()

    @x.setter
    def x(self, x: RealArray) -> None:
        """
        Update the values of the free DOFs with the supplied values

        Args:
            x: Array of new DOF values
               (word of caution: This setter blindly broadcasts a single value.
               So don't supply a single value unless you really desire.)
        """
        if len(self.free[self.free]) != len(x):
            raise ValueError  # To prevent fully fixed DOFs from not raising Error
        self.loc[self.free, "_x"] = x

    @property
    def full_x(self) -> RealArray:
        """
        Return all x even the fixed ones

        Returns:
            The values of full DOFs without any restrictions
        """
        return self._x.to_numpy()

    @property
    def reduced_len(self) -> Integral:
        """
        The number of free DOFs.

        The standard len function returns the full length of DOFs.

        Returns:
            The number of free DOFs
        """
        return len(self.free[self.free])

    @property
    def lower_bounds(self) -> RealArray:
        """
        Lower bounds of the DOFs

        Returns:
            Lower bounds of the DOFs
        """
        return self.loc[self.free, "_lb"].to_numpy()

    @lower_bounds.setter
    def lower_bounds(self, lower_bounds: RealArray) -> None:
        """

        Args:
            lower_bounds: Lower bounds of the DOFs
        """
        self.loc[self.free, "_lb"] = lower_bounds

    @property
    def upper_bounds(self) -> RealArray:
        """

        Returns:
            Upper bounds of the DOFs
        """
        return self.loc[self.free, "_ub"].to_numpy()

    @upper_bounds.setter
    def upper_bounds(self, upper_bounds: RealArray) -> None:
        """

        Args:
            upper_bounds: Upper bounds of the DOFs
        """
        self.loc[self.free, "_ub"] = upper_bounds

    @property
    def bounds(self) -> Tuple[RealArray, RealArray]:
        """

        Returns:
            (Lower bounds list, Upper bounds list)
        """
        return (self.lower_bounds, self.upper_bounds)

    def update_lower_bound(self, key: Key, val: Real) -> None:
        """
        Updates the lower bound of the specified DOF to the given value

        Args:
            key: DOF identifier
            val: Numeric lower bound of the DOF
        """
        if isinstance(key, str):
            self.loc[key, "_lb"] = val
        else:
            self.iloc[key, 2] = val

    def update_upper_bound(self, key: Key, val: Real) -> None:
        """
        Updates the upper bound of the specified DOF to the given value

        Args:
            key: DOF identifier
            val: Numeric upper bound of the DOF
        """
        if isinstance(key, str):
            self.loc[key, "_ub"] = val
        else:
            self.iloc[key, 3] = val

    def update_bounds(self, key: Key, val: Tuple[Real, Real]) -> None:
        """
        Updates the bounds of the specified DOF to the given value

        Args:
            key: DOF identifier
            val: (lower, upper) bounds of the DOF
        """
        if isinstance(key, str):
            self.loc[key, "_lb"] = val[0]
            self.loc[key, "_ub"] = val[1]
        else:
            self.iloc[key, 2] = val[0]
            self.iloc[key, 3] = val[1]

    @property
    def names(self):
        """

        Returns:
            string identifiers of the DOFs
        """
        return self.index.values


class Optimizable(ABC_Callable, Hashable, metaclass=OptimizableMeta):
    """
    Experimental callable ABC that provides lego-like optimizable objects
    that can be used to partition the optimization problem into a graph.

    The class provides many features that simplify defining the optimization
    problem.

    1. Optimizable and its subclasses define the optimization problem. The
       optimization problem can be thought of a directed acycling graph (DAG),
       which each instance of Optimizable being a vertex (node) in the DAG.
       Each Optimizable object can take other Optimizable objects as inputs and
       through this container logic, the edges of the DAG are defined.

       Alternatively, the input Optimizable objects can be thought as parents
       to the current Optimizable object. In this approach, the last grand-child
       defines the optimization problem by embodying all the elements of the
       parents and grand-parents.

       Each call to child instance gets in turn propagated to the parent. In this
       way, the last child acts as the final optimization problem to be solved.
       For an example of the final optimization node, refer to
       simsopt.objectives.least_squares.LeastSquaresProblem

    2. The class automatically partitions degrees of freedoms (DOFs) of
       the optimization problem to the associated Optimizable nodes. Each DOF
       defined in a parent gets passed down to the children as a needed DOF for
       the child. So a DOF needed by parent node can be given as an input to
       the methods in the child node. Any of the DOFs could be fixed in which
       case, it should be removed as an argument to the call-back
       function from the final Optimizable node.

    3. The class implements callable hook that provides minimal caching.
       All derived classes have to register methods that return objective function
       type values. This is done by implementing the following class attribute
       in the class definition:
       .. code-block:: python

           return_fn_map = {'name1': method1, 'name2': method, ...}

       The Optimizable class maintains the list of return functions needed by each
       of the calling Optimizable objects either during child initialization or
       later using the provided methods. The calling optimizable object could then
       call the Optimizable object directly using the `__call__` hook or could
       call the individual methods.

       This back and forth propagation of DOFs partitioning and function
       calls happens dynamically.

    4. The class is hashable and the names of the instances are unique. So
       instances of Optimizable class can be used as keys.

    Note:
        1. If the Optimizable object is called using the `__call__` hook, make
           sure to supply the argument `child=self`

        2. __init__ takes instances of subclasses of Optimizable as
           input and modifies them to add the current object as a child for
           input objects. The return fns of the parent object needed by the child
           could be specified by using `opt_return_fns` argument
    """
    return_fn_map: Dict[str, Callable] = NotImplemented

    def __init__(self,
                 x0: RealArray = None,
                 names: StrArray = None,
                 fixed: BoolArray = None,
                 lower_bounds: RealArray = None,
                 upper_bounds: RealArray = None,
                 opts_in: Sequence[Optimizable] = None,
                 opt_return_fns: Sequence[Sequence[str]] = None,
                 funcs_in: Sequence[Callable[..., Union[RealArray, Real]]] = None):
        """
        Args:
            x0: Initial state (or initial values of DOFs)
            names: Human identifiable names for the DOFs
            fixed: Array describing whether the DOFs are free or fixed
            lower_bounds: Lower bounds for the DOFs
            upper_bounds: Upper bounds for the DOFs
            opts_in: Sequence of Optimizable objects to define the optimization
                problem in conjuction with the DOFs. If the optimizable problem
                can be thought of as a direct acyclic graph based on
                dependencies, the optimizable objects
                supplied with opts_in act as parent nodes to the current
                Optimizable object in such an optimization graph
            opt_return_fns: Specifies the return value for each of the
                Optimizable object. Used in the case, where Optimizable object
                can return different return values. Typically return values are
                computed by different functions defined in the Optimizable
                object. The return values are selected by choosing the
                functions. To know the various return values, use the
                Optimizable.print_return_fn_names function. If the list is
                empty, default return value is used. If the Optimizable
                object can return multiple values, the default is the array
                of all possible return values.
            funcs_in: Instead of specifying opts_in and opt_return_fns, specify
                the methods of the Optimizable objects directly. The parent
                objects are identified automatically. Doesn't work with
                funcs_in with a property decorator
        """
        self._dofs = DOFs(x0,
                          names,
                          np.logical_not(fixed) if fixed is not None else None,
                          lower_bounds,
                          upper_bounds)

        # Generate unique and immutable representation for different
        # instances of same class
        self._id = ImmutableId(next(self.__class__._ids))
        self.name = self.__class__.__name__ + str(self._id.id)
        self._children = []  # This gets populated when the object is passed
        # as argument to another Optimizable object
        self.return_fns = defaultdict(list)  # Store return fn's required by each child

        # Assign self as child to parents
        self.parents = opts_in if opts_in is not None else []
        for i, parent in enumerate(self.parents):
            parent._add_child(self)
            return_fns = opt_return_fns[i] if opt_return_fns else []
            for fn in return_fns:
                parent.add_return_fn(self, fn)

        # Process funcs_in (Assumes opts_in is empty)
        if opts_in is None or not len(opts_in):
            opts_in = []
            funcs_in = funcs_in if funcs_in is not None else []
            for fn in funcs_in:
                opt_in = fn.__self__
                opts_in.append(opt_in)
                opt_in.add_return_fn(self, fn.__func__)
            opts_in = list(dict.fromkeys(opts_in))
            self.parents = list(opts_in) if opts_in is not None else []
            for i, parent in enumerate(self.parents):
                parent._add_child(self)

        # Obtain unique list of the ancestors
        self.ancestors = self._get_ancestors()

        # Compute the indices of all the DOFs
        self._update_free_dof_size_indices()
        self._update_full_dof_size_indices()

        self.new_x = True   # Set this True for dof setter and set it to False
        # after evaluation of function if True

    def __str__(self):
        return self.name

    def __hash__(self) -> int:
        hash_str = hashlib.sha256(self.name.encode('utf-8')).hexdigest()
        return int(hash_str, 16) % 10**32  # 32 digit int as hash

    def __eq__(self, other: Optimizable) -> bool:
        """
        Checks the equality condition

        Args:
            other: Another object of subclass of Optimizable

        Returns: True only if both are the same objects.

        """
        #return (self.__class__ == other.__class__ and
        #        self._id.id == other._id.id)
        return self.name == other.name

    def __call__(self, x: RealArray = None, *args, child=None, **kwargs):
        if x is not None:
            self.x = x
            self.new_x = True

        return_fn_map = self.__class__.return_fn_map
        if self.new_x:
            result = []
            for fn in return_fn_map.values():
                result.append(
                    fn(self, *args, **kwargs))
            self._val = result
            self.new_x = False

        if child:
            return_fns = self.return_fns[child] if self.return_fns[child] else \
                return_fn_map.values()
        else:
            return_fns = return_fn_map.values()

        return_result = []
        for fn in return_fns:
            i = list(return_fn_map.values()).index(fn)
            return_result.append(self._val[i])

        return return_result if len(return_result) > 1 else return_result[0]

    #@abc.abstractmethod
    #def f(self, *args, **kwargs):
    #    """
    #    Defines the callback function associated with the Optimizable subclass.

    #   Define the callback method in subclass of Optimizable. The function
    #    uses the state (x) stored self._dofs. To prevent hard-to-understand
    #    bugs, don't forget to use self.full_x.

    #    Args:
    #        *args:
    #        **kwargs:

    #    Returns:

    #    """

    def print_return_fn_names(self) -> List[str]:
        """
        Prints the names of the functions that could be used as objective
        functions.

        Returns:
            List of function names that could be used as objective functions
        """
        return list(self.__class__.return_fn_map.keys())

    def add_return_fn(self, child: Optimizable, fn: Union[str, Callable]) -> None:
        """
        Add return function to the list of the return functions called by
        the child Optimizable object

        Args:
            child: an Optimizable object that is direct dependent of the current
                Optimizable instance
            fn: method of the Optimizable object needed by the child
        """
        self._add_child(child)

        if isinstance(fn, str):
            fn = self.__class__.return_fn_map[fn]
        self.return_fns[child].append(fn)

    def get_return_fns(self, child: Optimizable) -> List[Callable]:
        """
        Gets return functions from this Optimizable object used by the child
        Optimizable object

        Args:
            child: Dependent Optimizable object

        Returns:
            List of methods that return a value when the current Optimizable
            object is called from the child
        """
        return self.return_fns.get(child, list(self.return_fn_map.values()))

    def get_return_fn_list(self) -> List[List[Callable]]:
        """
        Gets return functions from this Optimizable object used by all the child
        Optimizable objects

        Returns:
            List of methods that return a value when the current Optimizable
            object is called from the children.
        """
        # TODO: There could be a bug here
        return list(self.return_fns.values())

    def get_parent_return_fns_list(self) -> List[List[Callable]]:
        """
        Get a list of the funcs returned by the parents as list of lists

        Returns:
            The funcs returned by all the parents of the Optimizable object
        """
        return_fn_list = []
        for parent in self.parents:
            return_fn_list.append(parent.get_return_fns(self))
        return return_fn_list

    @property
    def parent_return_fns_no(self) -> int:
        """
        Compute the total number of the return funcs of all the parents
        of the Optimizable object

        Returns:
            The total number of the return funcs  of the Optimizable object's
            parents.
        """
        return_fn_no = 0
        for parent in self.parents:
            return_fn_no += len(parent.get_return_fns(self))
        return return_fn_no

    def _add_child(self, child: Optimizable) -> None:
        """
        Adds another Optimizable object as child. All the
        required processing of the dependencies is done in the child node.
        This method is used mainly to maintain 2-way link between parent
        and child.

        Args:
            child: Direct dependent (child) of the Optimizable object
        """
        if child not in self._children:
            self._children.append(child)

    def _remove_child(self, other: Optimizable) -> None:
        """
        Remove the specific Optimizable object from the children list.

        Args:
            child: Direct dependent (child) of the Optimizable object
        """
        self._children.remove(other)
        if other in self.return_fns:
            del self.return_fns[other]

    def add_parent(self, index: int, other: Optimizable) -> None:
        """
        Adds another Optimizable object as parent at specified index.

        Args:
            int: Index of the parent's list
            other: Another Optimizable object to be added as parent
        """
        if other not in self.parents:
            self.parents.insert(index, other)
            other._add_child(self)
            self.ancestors = self._get_ancestors()
            self._update_free_dof_size_indices()
            self._update_full_dof_size_indices()
            self.new_x = True
        else:
            print("The given Optimizable object is already a parent")

    def append_parent(self, other: Optimizable) -> None:
        """
        Appends another Optimizable object to parents list

        Args:
            other: New parent Optimizable object
        """
        if other not in self.parents:
            self.parents.append(other)
            other._add_child(self)
            self.ancestors = self._get_ancestors()
            self._update_free_dof_size_indices()
            self._update_full_dof_size_indices()
            self.new_x = True
        else:
            print("The given Optimizable object is already a parent")

    def pop_parent(self, index: int = -1) -> Optimizable:
        """
        Removes the parent Optimizable object at specified index.

        Args:
            index: Index of the list of the parents

        Returns:
            The removed parent Optimizable object
        """
        discarded_parent = self.parents.pop(index)
        discarded_parent._remove_child(self)
        self.ancestors = self._get_ancestors()
        self._update_free_dof_size_indices()
        self._update_full_dof_size_indices()
        self.new_x = True

        return discarded_parent

    def remove_parent(self, other: Optimizable):
        """
        Removes the specified Optimizable object from the list of parents.

        Args:
            other: The Optimizable object to be removed from the list of parents
        """
        self.parents.remove(other)
        other._remove_child(self)
        self.ancestors = self._get_ancestors()
        self._update_free_dof_size_indices()
        self._update_full_dof_size_indices()
        self.new_x = True

    @property
    def full_dof_size(self) -> Integral:
        """
        Total number of all (free and fixed) DOFs associated with the
        Optimizable object as well as parent Optimizable objects.
        """
        return self._full_dof_size

    @property
    def dof_size(self) -> Integral:
        """
        Total number of free DOFs associated with the Optimizable object
        as well as parent Optimizable objects.
        """
        return self._free_dof_size

    @property
    def local_full_dof_size(self) -> Integral:
        """
        Number of all (free and fixed) DOFs associated with the Optimizable
        object.

        Returns:
            Total number of free and fixed DOFs associated with the Optimizable
            object.
        """
        return len(self._dofs)

    @property
    def local_dof_size(self) -> Integral:
        """
        Number of free DOFs associated with the Optimizable object.

        Returns:
            Number of free DOFs associated with the Optimizable object.
        """
        return self._dofs.reduced_len

    def _update_free_dof_size_indices(self) -> None:
        """
        Updates the DOFs lengths for the Optimizable object as well as
        those of the descendent (dependent) Optimizable objects.

        Call this function whenever DOFs are fixed or unfixed or when parents
        are added/deleted. Recursively calls the same function in children
        """
        # TODO: This is slow because it walks through the graph repeatedly
        # TODO: Develop a faster scheme.
        # TODO: Alternatively ask the user to call this manually from the end
        # TODO: node after fixing/unfixing any DOF
        dof_indices = [0]
        free_dof_size = 0
        for opt in (self.ancestors + [self]):
            size = opt.local_dof_size
            free_dof_size += size
            dof_indices.append(free_dof_size)
        self._free_dof_size = free_dof_size
        self.dof_indices = dict(zip(self.ancestors + [self],
                                    zip(dof_indices[:-1], dof_indices[1:])))

        # Update the reduced length of children
        for child in self._children:
            child._update_free_dof_size_indices()

    def _update_full_dof_size_indices(self) -> None:
        """
        Updates the full DOFs lengths for this instance and
        those of the children.

        Call this function whenever parents are added or removed. Recursively
        calls the same function in children.
        """

        # TODO: This is slow because it walks through the graph repeatedly
        # TODO: Develop a faster scheme.
        # TODO: Alternatively ask the user to call this manually from the end
        # TODO: node after fixing/unfixing any DOF
        full_dof_size = 0
        for opt in (self.ancestors + [self]):
            full_dof_size += opt.local_full_dof_size
        self._full_dof_size = full_dof_size

        # Update the reduced length of children
        for child in self._children:
            child._update_full_dof_size_indices()

    @property
    def x(self) -> RealArray:
        """
        Numeric values of the free DOFs associated with the current
        Optimizable object and those of its ancestors
        """
        return np.concatenate([opt._dofs.x for
                               opt in (self.ancestors + [self])])

    @x.setter
    def x(self, x: RealArray) -> None:
        if list(self.dof_indices.values())[-1][-1] != len(x):
            raise ValueError
        for opt, indices in self.dof_indices.items():
            opt.local_x = x[indices[0]:indices[1]]
        self._set_new_x()

    @property
    def full_x(self) -> RealArray:
        """
        Numeric values of all the DOFs associated with the current
        Optimizable object and those of its ancestors
        """
        return np.concatenate([opt._dofs.full_x for
                               opt in (self.ancestors + [self])])

    @property
    def local_x(self) -> RealArray:
        """
        Numeric values of the free DOFs associated with this
        Optimizable object
        """
        return self._dofs.x

    @local_x.setter
    def local_x(self, x: RealArray) -> None:
        if self.local_dof_size != len(x):
            raise ValueError
        self._dofs.loc[self._dofs.free, '_x'] = x
        self.new_x = True

    @property
    def local_full_x(self):
        """
        Numeric values of all DOFs associated with this Optimizable object
        """
        return self._dofs.full_x

    def _set_new_x(self):
        self.new_x = True
        for child in self._children:
            child._set_new_x()

    @property
    def bounds(self) -> Tuple[RealArray, RealArray]:
        """
        Lower and upper bounds of the free DOFs associated with the current
        Optimizable object and those of its ancestors
        """
        return (self.lower_bounds, self.upper_bounds)

    @property
    def local_bounds(self) -> Tuple[RealArray, RealArray]:
        """
        Lower and upper bounds of the free DOFs associated with
        this Optimizable object
        """
        return self._dofs.bounds

    @property
    def lower_bounds(self) -> RealArray:
        """
        Lower bounds of the free DOFs associated with the current
        Optimizable object and those of its ancestors
        """
        opts = self.ancestors + [self]
        return np.concatenate([opt._dofs.lower_bounds for opt in opts])

    @property
    def local_lower_bounds(self) -> RealArray:
        """
        Lower bounds of the free DOFs associated with this Optimizable
        object
        """
        return self._dofs.lower_bounds

    @property
    def upper_bounds(self) -> RealArray:
        """
        Upper bounds of the free DOFs associated with the current
        Optimizable object and those of its ancestors
        """
        opts = self.ancestors + [self]
        return np.concatenate([opt._dofs.upper_bounds for opt in opts])

    @property
    def local_upper_bounds(self) -> RealArray:
        """
        Upper bounds of the free DOFs associated with this Optimizable
        object
        """
        return self._dofs.upper_bounds

    @local_upper_bounds.setter
    def local_upper_bounds(self, lub: RealArray) -> None:
        self._dofs.upper_bounds = lub

    @property
    def dof_names(self) -> StrArray:
        """
        Names (Identifiers) of the DOFs associated with the current
        Optimizable object and those of its ancestors
        """
        opts = self.ancestors + [self]
        return np.concatenate([opt._dofs.names for opt in opts])

    @property
    def local_dof_names(self) -> StrArray:
        """
        Names (Identifiers) of the DOFs associated with this Optimizable
        object
        """
        return self._dofs.names

    def get(self, key: Key) -> Real:
        """
        Get the value of specified DOF
        Even fixed dofs can be obtained individually.

        Args:
            key: DOF identifier
        """
        if isinstance(key, str):
            return self._dofs.loc[key, '_x']
        else:
            return self._dofs.iloc[key, 0]

    def set(self, key: Key, new_val: Real) -> None:
        """
        Update the value held the specified DOF.
        Even fixed dofs can be set this way

        Args:
            key: DOF identifier
            new_val: New value of the DOF
        """
        if isinstance(key, str):
            self._dofs.loc[key, '_x'] = new_val
        else:
            self._dofs.iloc[key, 0] = new_val

    @property
    def dofs_free_status(self) -> BoolArray:
        """
        Boolean array denoting whether the DOFs associated with the
        current and ancestors Optimizable objects are free or not
        """
        opts = self.ancestors + [self]
        return np.concatenate([opt._dofs.free.to_numpy() for opt in opts])

    @property
    def local_dofs_free_status(self) -> BoolArray:
        """
        Boolean array denoting whether the DOFs associated with the
        current Optimizable object are free or not
        """
        return self._dofs.free.to_numpy()

    def is_fixed(self, key: Key) -> bool:
        """
        Checks if the specified dof is fixed

        Args:
            key: DOF identifier
        """
        return not self.is_free(key)

    def is_free(self, key: Key) -> bool:
        """
        Checks if the specified dof is free

        Args:
            key: DOF identifier
        """
        if isinstance(key, str):
            return self._dofs.loc[key, 'free']
        else:
            return self._dofs.iloc[key, 1]

    def fix(self, key: Key) -> None:
        """
        Set the fixed attribute for the given degree of freedom.

        Args:
            key: DOF identifier
        """
        # TODO: Question: Should we use ifix similar to pandas' loc and iloc?
        # TODO: If key (str) is not found, it is silently ignored. Instead
        # TODO: raise a warning

        self._dofs.fix(key)
        self._update_free_dof_size_indices()

    def unfix(self, key: Key) -> None:
        """
        Unset the fixed attribute for the given degree of freedom

        Args:
            key: DOF identifier
        """
        # TODO: If key (str) is not found, it is silently ignored. Instead
        # TODO: raise a warning
        self._dofs.unfix(key)
        self._update_free_dof_size_indices()

    def fix_all(self) -> None:
        """
        Set the 'fixed' attribute for all degrees of freedom associated with
        the current Optimizable object.
        """
        #self.dof_fixed = np.full(len(self.get_dofs()), True)
        self._dofs.fix_all()
        self._update_free_dof_size_indices()

    def unfix_all(self) -> None:
        """
        Unset the 'fixed' attribute for all degrees of freedom associated
        with the current Optimizable object.
        """
        #self.dof_fixed = np.full(len(self.get_dofs()), False)
        self._dofs.unfix_all()
        self._update_free_dof_size_indices()

    def _get_ancestors(self) -> list[Optimizable]:
        """
        Get all the ancestors of the current Optimizable object

        Returns:
            List of Optimizable objects that are parents of current
            Optimizable objects
        """
        ancestors = []
        for parent in self.parents:
            ancestors += parent.ancestors
        ancestors += self.parents
        return list(dict.fromkeys(ancestors))


# TODO: Target class needs to be reimplemented to account for
# TODO: reimplementation of  Optimizable class
#class Target(Optimizable):
#    """
#    Given an attribute of an object, which typically would be a
#    @property, form a callable function that can be used as a target
#    for optimization.
#    """
#    def __init__(self, obj, attr):
#        self.obj = obj
#        self.attr = attr
#        super().__init__()
#
#        # Attach a dJ function only if obj has one
#        def dJ(self0):
#            return getattr(self0.obj, 'd' + self0.attr)
#
#        if hasattr(obj, 'd' + attr):
#            self.dJ = types.MethodType(dJ, self)
#
#    def f(self):
#        # TODO: Implemnt the f to call self.obj.attr
#        return getattr(self.obj, self.attr)
#
#    @deprecated(version='0.0.2', reason="Call the object directly. Don't assume"
#                                        " J method will be present.")
#    def J(self):
#        return getattr(self.obj, self.attr)
#
#    #def dJ(self):
#    #    return getattr(self.obj, 'd' + self.attr)
#
#    # Bharat's comment: The following two needs to be better defined
#    def get_dofs(self):
#        return np.array([])
#
#    def set_dofs(self, v):
#        pass
#
#
#def function_from_user(target):
#    """
#    Given a user-supplied "target" to be optimized, extract the
#    associated callable function.
#    """
#    if callable(target):
#        return target
#    elif hasattr(target, 'J') and callable(target.J):
#        return target.J
#    else:
#        raise TypeError('Unable to find a callable function associated '
#                        'with the user-supplied target ' + str(target))
#
## TODO: make_optimizable function should be reimplemented to account for
## TODO: reimplementation of Optimizable class
#def make_optimizable(obj):
#    """
#    Given any object, add attributes like fixed, mins, and maxs. fixed
#    = False by default. Also, add the other methods of Optimizable to
#    the object.
#    """
#
#    # If the object does not have a get_dofs() method, attach one,
#    # assuming the object does not directly own any dofs.
#    def get_dofs(self):
#        return np.array([])
#    def set_dofs(self, x):
#        pass
#    if not hasattr(obj, 'get_dofs'):
#        obj.get_dofs = types.MethodType(get_dofs, obj)
#    if not hasattr(obj, 'set_dofs'):
#        obj.set_dofs = types.MethodType(set_dofs, obj)
#
#    n = len(obj.get_dofs())
#    if not hasattr(obj, 'dof_fixed'):
#        obj.dof_fixed = np.full(n, False)
#    if not hasattr(obj, 'mins'):
#        obj.mins = np.full(n, np.NINF)
#    if not hasattr(obj, 'maxs'):
#        obj.maxs = np.full(n, np.Inf)
#
#    # Add the following methods from the Optimizable class:
#    #for method in ['index', 'get', 'set', 'get_fixed', 'set_fixed', 'all_fixed']:
#    # See https://stackoverflow.com/questions/972/adding-a-method-to-an-existing-object-instance
#    #setattr(obj, method, types.MethodType(getattr(Optimizable, method), obj))
#
#    # New compact implementation
#    method_list = [f for f in dir(Optimizable) if \
#            callable(getattr(Optimizable, f)) and not f.startswith("__")]
#    for f in method_list:
#        if not hasattr(obj, f) and f not in ('get_dofs', 'set_dofs'):
#            setattr(obj, f, types.MethodType(getattr(Optimizable, f), obj))
#
#    return obj



import numpy as np


def draw_uniform_on_curve(curve, nsamples, safetyfactor=10):
    r"""
    Uses rejection sampling to sample points on a curve. *Warning*: assumes that
    the underlying quadrature points on the Curve are uniformly distributed.

    Args:
        curve: The :mod:`simsopt.geo.curve.Curve` to spawn the particles on.
        nsamples: number of samples.
        safetyfactor: how many more samples than ``nsamples`` to generate for
                      rejection/acceptance.
    """
    alen = curve.incremental_arclength()
    M = np.max(alen)
    nattempts = 10 * nsamples
    idxs = np.random.randint(0, alen.shape[0], size=(nattempts, ))
    accept = np.where(np.random.uniform(low=0, high=1, size=(nattempts, )) < alen[idxs]/M)[0]
    assert len(accept) > nsamples
    idxs = np.sort(idxs[accept[:nsamples]])
    xyz = curve.gamma()[idxs, :]
    return xyz, idxs


def draw_uniform_on_surface(surface, nsamples, safetyfactor=10):
    r"""
    Uses rejection sampling to sample points on a surface. *Warning*: assumes that
    the underlying quadrature points on the surface are uniformly distributed.

    Args:
        surface: The :mod:`simsopt.geo.surface.Surface` to spawn the particles
                 on.
        nsamples: number of samples.
        safetyfactor: how many more samples than ``nsamples`` to generate for
                      rejection/acceptance.
    """
    jac = np.linalg.norm(surface.normal().reshape((-1, 3)), axis=1)
    M = np.max(jac)
    nattempts = 10 * nsamples
    idxs = np.random.randint(0, jac.shape[0], size=(nattempts, ))
    accept = np.where(np.random.uniform(low=0, high=1, size=(nattempts, )) < jac[idxs]/M)[0]
    assert len(accept) > nsamples
    idxs = np.sort(idxs[accept[:nsamples]])
    gamma = surface.gamma()
    order = 'F' if np.isfortran(gamma) else 'C'
    idxs = np.unravel_index(idxs, gamma.shape[:2], order)
    xyz = gamma[idxs[0], idxs[1], :]
    return xyz, idxs



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module provides a class that handles the VMEC equilibrium code.
"""

import logging
import os.path
from typing import Union

import numpy as np
from scipy.io import netcdf
from ..util.dev import SimsoptRequires

logger = logging.getLogger(__name__)

try:
    from mpi4py import MPI
except ImportError as e:
    MPI = None 
    logger.warning(str(e))

# vmec_found = True
try:
    import vmec
except ImportError as e:
    # vmec_found = False
    vmec = None
    logger.warning(str(e))

from .._core.optimizable import Optimizable
from .._core.util import Struct, ObjectiveFailure
from ..geo.surfacerzfourier import SurfaceRZFourier

if MPI is not None:
    from ..util.mpi import MpiPartition
else:
    MpiPartition = None


# Flags used by runvmec():
restart_flag = 1
readin_flag = 2
timestep_flag = 4
output_flag = 8
cleanup_flag = 16
reset_jacdt_flag = 32

# Documentation of flags for runvmec() from the VMEC source code:
#
#value flag-name         calls routines to...
#----- ---------         ---------------------
#  1   restart_flag      reset internal run-control parameters
#                        (for example, if jacobian was bad, to try a smaller 
#                        time-step)
#  2   readin_flag       read in data from input_file and initialize parameters
#                        or arrays which do not dependent on radial grid size
#                        allocate internal grid-dependent arrays used by vmec;
#                        initialize internal grid-dependent vmec profiles (xc,
#                        iota, etc);
#                        setup loop for radial multi-grid meshes or, if
#                        ns_index = ictrl_array(4) is > 0, use radial grid
#                        points specified by ns_array[ns_index]
#  4   timestep_flag     iterate vmec either by "niter" time steps or until ftol
#                        satisfied, whichever comes first.
#                        If numsteps (see below) > 0, vmec will return
#                        to caller after numsteps, rather than niter, steps.
#  8   output_flag       write out output files (wout, jxbout)
# 16   cleanup_flag      cleanup (deallocate arrays) - this terminates present
#                        run of the sequence
#                        This flag will be ignored if the run might be continued.
#                        For example, if ier_flag (see below) returns the value
#                        more_iter_flag, the cleanup code will be skipped even if
#                        cleanup_flag is set, so that the run could be continued
#                        on the next call to runvmec.
# 32   reset_jacdt_flag  Resets ijacobian flag and time step to delt0
#                        thus, setting ictrl_flag = 1+2+4+8+16 will perform ALL
#                        the tasks thru cleanup_flag in addition,
#                        if ns_index = 0 and numsteps = 0 (see below), vmec will
#                        control its own run history


# Temporarily commenting out the decorator till __instancecheck__ method is made working
#@SimsoptRequires(MPI is not None, "mpi4py needs to be installed for running VMEC")
class Vmec(Optimizable):
    """
    This class represents the VMEC equilibrium code.

    The input parameters to VMEC are all accessible as attributes of
    the ``indata`` attribute. For example, if ``vmec`` is an instance
    of ``Vmec``, then you can read or write the input resolution
    parameters using ``vmec.indata.mpol``, ``vmec.indata.ntor``,
    ``vmec.indata.ns_array``, etc. However, the boundary surface is
    different: ``rbc``, ``rbs``, ``zbc``, and ``zbs`` from the
    ``indata`` attribute are always ignored, and these arrays are
    instead taken from the simsopt surface object associated to the
    ``boundary`` attribute. If ``boundary`` is a surface based on some
    other representation than VMEC's Fourier representation, the
    surface will automatically be converted to VMEC's representation
    (:obj:`~simsopt.geo.surfacerzfourier.SurfaceRZFourier`) before
    each run of VMEC. You can replace ``boundary`` with a new surface
    object, of any type that implements the conversion function
    ``to_RZFourier()`.

    VMEC is run either when the :meth:`run()` function is called, or when
    any of the output functions like :meth:`aspect()` or :meth:`iota_axis()`
    are called.

    A caching mechanism is implemented, using the attribute
    ``need_to_run_code``. Whenever VMEC is run, this attribute is set
    to ``False``. Subsequent calls to :meth:`run()` or output
    functions like :meth:`aspect()` will not actually run VMEC again,
    until ``need_to_run_code`` is changed to ``True``. The attribute
    ``need_to_run_code`` is automatically set to ``True`` whenever
    :meth:`set_dofs()` is called. However, ``need_to_run_code`` is not
    automatically set to ``True`` when entries of ``indata`` are
    modified, or when ``boundary`` is modified.

    Once VMEC has run at least once, all of the quantities in the
    ``wout`` output file are available as attributes of the ``wout``
    attribute.  For example, if ``vmec`` is an instance of ``Vmec``,
    then the flux surface shapes can be obtained from
    ``vmec.wout.rmnc`` and ``vmec.wout.zmns``.

    Since the underlying fortran implementation of VMEC uses global
    module variables, it is not possible to have more than one python
    Vmec object with different parameters; changing the parameters of
    one would change the parameters of the other.

    An instance of this class owns just a few optimizable degrees of
    freedom, particularly ``phiedge`` and ``curtor``. The optimizable
    degrees of freedom associated with the boundary surface are owned
    by that surface object.

    The default behavior is that all ``wout`` output files will be
    deleted except for the first and most recent iteration on worker
    group 0. If you wish to keep all the ``wout`` files, you can set
    ``keep_all_files = True``. If you want to save the ``wout`` file
    for a certain intermediate iteration, you can set the
    ``files_to_delete`` attribute to ``[]`` after that run of VMEC.

    Args:
        filename: Name of a VMEC input file to use for loading the 
          initial parameters. If ``None``, default parameters will be used.
        mpi: A :obj:`simsopt.util.mpi.MpiPartition` instance, from which 
          the worker groups will be used for VMEC calculations. If ``None``,
          each MPI process will run VMEC independently.
        keep_all_files: If ``False``, all ``wout`` output files will be deleted
          except for the first and most recent ones from worker group 0. If 
          ``True``, all ``wout`` files will be kept.

    Attributes:
        iter: Number of times VMEC has run.
        s_full_grid: The "full" grid in the radial coordinate s (normalized
          toroidal flux), including points at s=0 and s=1. Used for the output
          arrays and ``zmns``.
        s_half_grid: The "half" grid in the radial coordinate s, used for
          ``bmnc``, ``lmns``, and other output arrays. In contrast to
          wout files, this array has only ns-1 entries, so there is no
          leading 0.
        ds: The spacing between grid points for the radial coordinate s.
    """

    def __init__(self,
                 filename: Union[str, None] = None,
                 mpi: Union[MpiPartition, None] = None,
                 keep_all_files: bool = False):
        if MPI is None:
            raise RuntimeError("mpi4py needs to be installed for running VMEC")
        if vmec is None:
            raise RuntimeError(
                "Running VMEC from simsopt requires VMEC python extension. "
                "Install the VMEC python extension from "
                "https://https://github.com/hiddenSymmetries/VMEC2000")

        if filename is None:
            # Read default input file, which should be in the same
            # directory as this file:
            filename = os.path.join(os.path.dirname(__file__), 'input.default')
            logger.info("Initializing a VMEC object from defaults in "
                        + filename)
        else:
            logger.info("Initializing a VMEC object from file: " + filename)
        self.input_file = filename

        # Get MPI communicator:
        if mpi is None:
            self.mpi = MpiPartition(ngroups=1)
        else:
            self.mpi = mpi
        comm = self.mpi.comm_groups
        self.fcomm = comm.py2f()

        self.ictrl = np.zeros(5, dtype=np.int32)
        self.iter = -1
        self.keep_all_files = keep_all_files
        self.files_to_delete = []
        self.wout = Struct()

        self.ictrl[0] = restart_flag + readin_flag
        self.ictrl[1] = 0  # ierr
        self.ictrl[2] = 0  # numsteps
        self.ictrl[3] = 0  # ns_index
        self.ictrl[4] = 0  # iseq
        verbose = True
        reset_file = ''
        logger.info('About to call runvmec to readin')
        vmec.runvmec(self.ictrl, filename, verbose, self.fcomm, reset_file)
        ierr = self.ictrl[1]
        logger.info('Done with runvmec. ierr={}. Calling cleanup next.'.format(ierr))
        # Deallocate arrays allocated by VMEC's fixaray():
        vmec.cleanup(False)
        if ierr != 0:
            raise RuntimeError("Failed to initialize VMEC from input file {}. "
                               "error code {}".format(filename, ierr))

        objstr = " for Vmec " + str(hex(id(self)))

        # Create an attribute for each VMEC input parameter in VMEC's fortran
        # modules,
        self.indata = vmec.vmec_input  # Shorthand
        vi = vmec.vmec_input  # Shorthand
        # A vmec object has mpol and ntor attributes independent of
        # the boundary. The boundary surface object is initialized
        # with mpol and ntor values that match those of the vmec
        # object, but the mpol/ntor values of either the vmec object
        # or the boundary surface object can be changed independently
        # by the user.
        self.boundary = SurfaceRZFourier(nfp=vi.nfp,
                                         stellsym=not vi.lasym,
                                         mpol=vi.mpol,
                                         ntor=vi.ntor)
        self.free_boundary = bool(vi.lfreeb)

        # Transfer boundary shape data from fortran to the ParameterArray:
        for m in range(vi.mpol + 1):
            for n in range(-vi.ntor, vi.ntor + 1):
                self.boundary.rc[m, n + vi.ntor] = vi.rbc[101 + n, m]
                self.boundary.zs[m, n + vi.ntor] = vi.zbs[101 + n, m]
        # Handle a few variables that are not Parameters:
        self.depends_on = ["boundary"]
        self.need_to_run_code = True

        self.fixed = np.full(len(self.get_dofs()), True)
        self.names = ['delt', 'tcon0', 'phiedge', 'curtor', 'gamma']

    def get_dofs(self):
        return np.array([self.indata.delt, self.indata.tcon0,
                         self.indata.phiedge, self.indata.curtor,
                         self.indata.gamma])

    def set_dofs(self, x):
        self.need_to_run_code = True
        self.indata.delt = x[0]
        self.indata.tcon0 = x[1]
        self.indata.phiedge = x[2]
        self.indata.curtor = x[3]
        self.indata.gamma = x[4]

    def run(self):
        """
        Run VMEC, if ``need_to_run_code`` is ``True``.
        """
        if not self.need_to_run_code:
            logger.info("run() called but no need to re-run VMEC.")
            return
        logger.info("Preparing to run VMEC.")
        # Transfer values from Parameters to VMEC's fortran modules:
        vi = vmec.vmec_input  # Shorthand
        # Convert boundary to RZFourier if needed:
        boundary_RZFourier = self.boundary.to_RZFourier()
        # VMEC does not allow mpol or ntor above 101:
        if vi.mpol > 101:
            raise ValueError("VMEC does not allow mpol > 101")
        if vi.ntor > 101:
            raise ValueError("VMEC does not allow ntor > 101")
        vi.rbc[:, :] = 0
        vi.zbs[:, :] = 0
        mpol_capped = np.min([boundary_RZFourier.mpol, 101])
        ntor_capped = np.min([boundary_RZFourier.ntor, 101])
        # Transfer boundary shape data from the surface object to VMEC:
        for m in range(mpol_capped + 1):
            for n in range(-ntor_capped, ntor_capped + 1):
                vi.rbc[101 + n, m] = boundary_RZFourier.get_rc(m, n)
                vi.zbs[101 + n, m] = boundary_RZFourier.get_zs(m, n)

        # Set axis shape to something that is obviously wrong (R=0) to
        # trigger vmec's internal guess_axis.f to run. Otherwise the
        # initial axis shape for run N will be the final axis shape
        # from run N-1, which makes VMEC results depend slightly on
        # the history of previous evaluations, confusing the finite
        # differencing.
        vi.raxis_cc[:] = 0
        vi.raxis_cs[:] = 0
        vi.zaxis_cc[:] = 0
        vi.zaxis_cs[:] = 0

        self.iter += 1
        input_file = self.input_file + '_{:03d}_{:06d}'.format(
            self.mpi.group, self.iter)
        self.output_file = os.path.join(
            os.getcwd(),
            os.path.basename(input_file).replace('input.', 'wout_') + '.nc')
        mercier_file = os.path.join(
            os.getcwd(),
            os.path.basename(input_file).replace('input.', 'mercier.'))
        jxbout_file = os.path.join(
            os.getcwd(),
            os.path.basename(input_file).replace('input.', 'jxbout_') + '.nc')

        # I should write an input file here.
        logger.info("Calling VMEC reinit().")
        vmec.reinit()

        logger.info("Calling runvmec().")
        self.ictrl[0] = restart_flag + reset_jacdt_flag \
            + timestep_flag + output_flag
        self.ictrl[1] = 0  # ierr
        self.ictrl[2] = 0  # numsteps
        self.ictrl[3] = 0  # ns_index
        self.ictrl[4] = 0  # iseq
        verbose = True
        reset_file = ''
        vmec.runvmec(self.ictrl, input_file, verbose, self.fcomm, reset_file)
        ierr = self.ictrl[1]

        # Deallocate arrays, even if vmec did not converge:
        logger.info("Calling VMEC cleanup().")
        vmec.cleanup(True)

        # See VMEC2000/Sources/General/vmec_params.f for ierr codes.
        # 11 = successful_term_flag.
        # Error codes that are expected to occur due to lack of
        # convergence cause ObjectiveFailure, which the optimizer
        # handles gracefully by treating the point as bad. But the
        # user/developer should know if an error codes arises that
        # should logically never occur, so these codes raise a
        # different exception.
        if ierr in [0, 5]:
            raise RuntimeError(f"runvmec returned an error code that should " \
                               "never occur: ierr={ierr}")
        if ierr != 11:
            raise ObjectiveFailure(f"VMEC did not converge. ierr={ierr}")

        logger.info("VMEC run complete. Now loading output.")
        self.load_wout()
        logger.info("Done loading VMEC output.")

        # Group leaders handle deletion of files:
        if self.mpi.proc0_groups:
            # Delete some files produced by VMEC that we never care
            # about. For some reason the os.remove statements give a 'file
            # not found' error in the CI, hence the try-except blocks.
            try:
                os.remove(mercier_file)
            except FileNotFoundError:
                logger.debug(f'Tried to delete the file {mercier_file} but it was not found')
                raise

            try:
                os.remove(jxbout_file)
            except FileNotFoundError:
                logger.debug(f'Tried to delete the file {jxbout_file} but it was not found')
                raise

            try:
                os.remove("fort.9")
            except FileNotFoundError:
                logger.debug('Tried to delete the file fort.9 but it was not found')

            # If the worker group is not 0, delete all wout files, unless
            # keep_all_files is True:
            if (not self.keep_all_files) and (self.mpi.group > 0):
                os.remove(self.output_file)

            # Delete the previous output file, if desired:
            for filename in self.files_to_delete:
                os.remove(filename)
            self.files_to_delete = []

            # Record the latest output file to delete if we run again:
            if (self.mpi.group == 0) and (self.iter > 0) and (not self.keep_all_files):
                self.files_to_delete.append(self.output_file)

        self.need_to_run_code = False

    def load_wout(self):
        """
        Read in the most recent ``wout`` file created, and store all the
        data in a ``wout`` attribute of this Vmec object.
        """
        ierr = 0
        logger.info("Attempting to read file " + self.output_file)
        # vmec.read_wout_mod.read_wout_file(self.output_file, ierr)
        #if ierr == 0:
        #    logger.info('Successufully load VMEC results from ' + \
        #                self.output_file)
        # else:
        #    print('Load VMEC results from {:} failed!'.format(
        #        self.output_file))
        #wout = vmec.read_wout_mod
        #print('xm:', wout.xm)
        #print('xn:', wout.xn)
        #print('xm_nyq:', wout.xm_nyq)
        #print('xn_nyq:', wout.xn_nyq)
        #print('type(xm):', type(wout.xm), ' type(xn):', type(wout.xn), ' type(xm_nyq):', type(wout.xm_nyq), ' type(xn_nyq):', type(wout.xn_nyq))
        #print('ierr:', ierr)
        #print('mnmax:', wout.mnmax, ' len(xm):', len(wout.xm), ' len(xn):', len(wout.xn))
        #print('mnmax_nyq:', wout.mnmax_nyq, ' len(xm_nyq):', len(wout.xm_nyq), ' len(xn_nyq):', len(wout.xn_nyq))
        #assert len(wout.xm) == wout.mnmax
        #assert len(wout.xn) == wout.mnmax
        #assert len(wout.xm_nyq) == wout.mnmax_nyq
        #assert len(wout.xn_nyq) == wout.mnmax_nyq

        f = netcdf.netcdf_file(self.output_file, mmap=False)
        for key, val in f.variables.items():
            # 2D arrays need to be transposed.
            val2 = val[()]  # Convert to numpy array
            val3 = val2.T if len(val2.shape) == 2 else val2
            self.wout.__setattr__(key, val3)

        #self.wout.ier_flag = f.variables['ier_flag'][()]
        if self.wout.ier_flag != 0:
            logger.info("VMEC did not succeed!")
            raise ObjectiveFailure("VMEC did not succeed")

        # Shorthand for a long variable name:
        self.wout.lasym = f.variables['lasym__logical__'][()]
        self.wout.volume = self.wout.volume_p

        #self.wout.ier_flag = f.variables['ier_flag'][()]
        #if self.wout.ier_flag != 0:
        #    logger.info("VMEC did not succeed!")
        #    raise ObjectiveFailure("VMEC did not succeed")
        #self.wout.nfp = f.variables['nfp'][()]
        #self.wout.lasym = f.variables['lasym__logical__'][()]
        #self.wout.ns = f.variables['ns'][()]
        #self.wout.mnmax = f.variables['mnmax'][()]
        #self.wout.mnmax_nyq = f.variables['mnmax_nyq'][()]
        #self.wout.xm = f.variables['xm'][()]
        #self.wout.xn = f.variables['xn'][()]
        #self.wout.xm_nyq = f.variables['xm_nyq'][()]
        #self.wout.xn_nyq = f.variables['xn_nyq'][()]
        #self.wout.mpol = f.variables['mpol'][()]
        #self.wout.ntor = f.variables['ntor'][()]
        #self.wout.bmnc = f.variables['bmnc'][()].transpose()
        #self.wout.rmnc = f.variables['rmnc'][()].transpose()
        #self.wout.zmns = f.variables['zmns'][()].transpose()
        #self.wout.lmns = f.variables['lmns'][()].transpose()
        #self.wout.bsubumnc = f.variables['bsubumnc'][()].transpose()
        #self.wout.bsubvmnc = f.variables['bsubvmnc'][()].transpose()
        #self.wout.iotas = f.variables['iotas'][()]
        #self.wout.iotaf = f.variables['iotaf'][()]
        #self.wout.aspect = f.variables['aspect'][()]
        #self.wout.volume = f.variables['volume_p'][()]
        f.close()

        self.s_full_grid = np.linspace(0, 1, self.wout.ns)
        self.ds = self.s_full_grid[1] - self.s_full_grid[0]
        self.s_half_grid = self.s_full_grid[1:] - 0.5 * self.ds

        return ierr

    def aspect(self):
        """
        Return the plasma aspect ratio.
        """
        self.run()
        return self.wout.aspect

    def volume(self):
        """
        Return the volume inside the VMEC last closed flux surface.
        """
        self.run()
        return self.wout.volume

    def iota_axis(self):
        """
        Return the rotational transform on axis
        """
        self.run()
        return self.wout.iotaf[0]

    def iota_edge(self):
        """
        Return the rotational transform at the boundary
        """
        self.run()
        return self.wout.iotaf[-1]

    def mean_iota(self):
        """
        Return the mean rotational transform. The average is taken over
        the normalized toroidal flux s.
        """
        self.run()
        return np.mean(self.wout.iotas[1:])

    def mean_shear(self):
        """
        Return an average magnetic shear, d(iota)/ds, where s is the
        normalized toroidal flux. This is computed by fitting the
        rotational transform to a linear (plus constant) function in
        s. The slope of this fit function is returned.
        """
        self.run()

        # Fit a linear polynomial:
        poly = np.polynomial.Polynomial.fit(self.s_half_grid,
                                            self.wout.iotas[1:], deg=1)
        # Return the slope:
        return poly.deriv()(0)

    def get_max_mn(self):
        """
        Look through the rbc and zbs data in fortran to determine the
        largest m and n for which rbc or zbs is nonzero.
        """
        max_m = 0
        max_n = 0
        for m in range(1, 101):
            for n in range(1, 101):
                if np.abs(vmec.vmec_input.rbc[101 + n, m]) > 0 \
                        or np.abs(vmec.vmec_input.zbs[101 + n, m]) > 0 \
                        or np.abs(vmec.vmec_input.rbs[101 + n, m]) > 0 \
                        or np.abs(vmec.vmec_input.zbc[101 + n, m]) > 0 \
                        or np.abs(vmec.vmec_input.rbc[101 - n, m]) > 0 \
                        or np.abs(vmec.vmec_input.zbs[101 - n, m]) > 0 \
                        or np.abs(vmec.vmec_input.rbs[101 - n, m]) > 0 \
                        or np.abs(vmec.vmec_input.zbc[101 - n, m]) > 0:
                    max_m = np.max((max_m, m))
                    max_n = np.max((max_n, n))
        # It may happen that mpol or ntor exceed the max_m or max_n
        # according to rbc/zbs. In this case, go with the larger
        # value.
        max_m = np.max((max_m, vmec.vmec_input.mpol))
        max_n = np.max((max_n, vmec.vmec_input.ntor))
        return (max_m, max_n)

    def __repr__(self):
        """
        Print the object in an informative way.
        """
        return "Vmec instance " + str(hex(id(self))) \
            + " (nfp=" + str(self.indata.nfp) \
            + " mpol=" + str(self.indata.mpol) \
            + " ntor=" + str(self.indata.ntor) + ")"

    def vacuum_well(self):
        """
        Compute a single number W that summarizes the vacuum magnetic well,
        given by the formula

        W = (dV/ds(s=0) - dV/ds(s=1)) / (dV/ds(s=0)

        where dVds is the derivative of the flux surface volume with
        respect to the radial coordinate s. Positive values of W are
        favorable for stability to interchange modes. This formula for
        W is motivated by the fact that

        d^2 V / d s^2 < 0

        is favorable for stability. Integrating over s from 0 to 1
        and normalizing gives the above formula for W. Notice that W
        is dimensionless, and it scales as the square of the minor
        radius. To compute dV/ds, we use

        dV/ds = 4 * pi**2 * abs(sqrt(g)_{0,0})

        where sqrt(g) is the Jacobian of (s, theta, phi) coordinates,
        computed by VMEC in the gmnc array, and _{0,0} indicates the
        m=n=0 Fourier component. Since gmnc is reported by VMEC on the
        half mesh, we extrapolate by half of a radial grid point to s
        = 0 and 1.
        """

        self.run()

        # gmnc is on the half mesh, so drop the 0th radial entry:
        dVds = 4 * np.pi * np.pi * np.abs(self.wout.gmnc[0, 1:])

        # To get from the half grid to s=0 and s=1, we must
        # extrapolate by 1/2 of a radial grid point:
        dVds_s0 = 1.5 * dVds[0] - 0.5 * dVds[1]
        dVds_s1 = 1.5 * dVds[-1] - 0.5 * dVds[-2]

        well = (dVds_s0 - dVds_s1) / dVds_s0
        return well



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module provides the LeastSquaresProblem class, as well as the
associated class LeastSquaresTerm.
"""

__all__ = ['LeastSquaresTerm', 'LeastSquaresProblem']

import logging
import warnings
from typing import Union

import numpy as np
from scipy.optimize import least_squares

from .._core.dofs import Dofs
from .._core.util import isnumber
from .._core.optimizable import function_from_user, Target

logger = logging.getLogger(__name__)


class LeastSquaresTerm:
    """
    This class represents one term in a nonlinear-least-squares
    problem. A LeastSquaresTerm instance has 3 basic attributes: a
    function (called f_in), a goal value (called goal), and a weight
    (sigma).  The overall value of the term is:

    f_out = weight * (f_in - goal) ** 2.
    """

    def __init__(self, f_in, goal, weight):  # =None, sigma=None):
        self.f_in = function_from_user(f_in)
        self.goal = goal
        if not isnumber(weight):  # Bharat's comment: Do we need this check?
            raise TypeError('Weight must be a float or int')
        if weight < 0:
            raise ValueError('Weight cannot be negative')
        self.weight = float(weight)
        # self.fixed = np.full(0, False) # What is this line for?

    @classmethod
    def from_sigma(cls, f_in, goal, sigma):
        """
        Define the LeastSquaresTerm with sigma = 1 / sqrt(weight), so

        f_out = ((f_in - goal) / sigma) ** 2.
        """
        if sigma == 0:
            raise ValueError('sigma cannot be 0')
        if not isnumber(sigma):  # Bharat's comment: Do we need this check?
            raise TypeError('sigma must be a float or int')
        return cls(f_in, goal, 1.0 / float(sigma * sigma))

    def f_out(self):
        """
        Return the overall value of this least-squares term.
        """
        temp = self.f_in() - self.goal
        # Below, np.dot works with both scalars and vectors.
        return self.weight * np.dot(temp, temp)


class LeastSquaresProblem:
    """
    This class represents a nonlinear-least-squares optimization
    problem. The class stores a list of LeastSquaresTerm objects.

    Args:
        terms: Must be convertable to a list by the list()
          subroutine. Each entry of the resulting list must either have
          type LeastSquaresTerm or else be a list or tuple of the form
          (function, goal, weight) or (object, attribute_str, goal,
          weight).
        kwargs: Any additional arguments will be passed to the
          :class:`~simsopt._core.dofs.Dofs` constructor. This is useful for
          passing ``fail``, ``abs_step``, ``rel_step``, and ``diff_method``.
          See :class:`~simsopt._core.dofs.Dofs` for details.
    """

    def __init__(self,
                 terms,
                 **kwargs):

        self.kwargs = kwargs
        # For each item provided in the list, either convert to a
        # LeastSquaresTerm or, if it is already a LeastSquaresTerm,
        # use it directly.
        self.terms = []
        msg = 'Each term must be either (1) a LeastSquaresTerm or (2) a list '\
              'or tuple of the form (function, goal, weight) or (object, '    \
              'attribute_str, goal, weight)'
        for term in terms:
            if isinstance(term, LeastSquaresTerm):
                self.terms.append(term)
            else:  # Expect the term to be an Iterable, but don't check
                if len(term) == 4:  # 4 item list is a special case
                    lst = LeastSquaresTerm(Target(*term[:2]), *term[2:])
                else:
                    lst = LeastSquaresTerm(*term)
                self.terms.append(lst)

        if not len(self.terms):
            raise ValueError("At least 1 LeastSquaresTerm must be as argument")

        self._init()

    def _init(self):
        """
        Call collect_dofs() on the list of terms to set x, mins, maxs, names,
        etc. This is done both when the object is created, so 'objective' 
        works immediately, and also at the start of solve()
        """
        self.dofs = Dofs([t.f_in for t in self.terms], **self.kwargs)

    @property
    def x(self):
        """
        Return the state vector.
        """
        # Delegate to Dofs:
        return self.dofs.x

    @x.setter
    def x(self, x):
        """
        Sets the global state vector to x.
        """
        # Delegate to Dofs:
        if x is not None:
            self.dofs.set(x)

    def objective(self, x=None):
        """
        Return the value of the total objective function, by summing
        the terms.

        If the argument x is not supplied, the objective will be
        evaluated for the present state vector. If x is supplied, then
        first set_dofs() will be called for each object to set the
        global state vector to x.
        """
        logger.info("objective() called with x=" + str(x))
        self.x = x

        return self.objective_from_shifted_f(self.f())

    def f_from_unshifted(self, f_unshifted):
        """
        This function takes a vector of function values, as returned by
        dofs, and shifts and scales them. This function does not
        actually evaluate the dofs.
        """

        residuals = np.zeros(len(f_unshifted))
        start_index = 0
        for j in range(self.dofs.nfuncs):
            term = self.terms[j]
            end_index = start_index + self.dofs.nvals_per_func[j]
            residuals[start_index:end_index] = \
                (f_unshifted[start_index:end_index] - term.goal) * \
                np.sqrt(term.weight)
            start_index = end_index

        return residuals

    def f(self, x=None):
        """
        This method returns the vector of residuals for a given state
        vector x.  This function is passed to scipy.optimize, and
        could be passed to other optimization algorithms too.  This
        function differs from Dofs.f() because it shifts and scales
        the terms.

        If the argument x is not supplied, the residuals will be
        evaluated for the present state vector. If x is supplied, then
        first set_dofs() will be called for each object to set the
        global state vector to x.
        """
        logger.info("f() called with x=" + str(x))
        self.x = x

        # Importantly for MPI, the next line calls the functions in
        # the same order that Dofs.f() does. Proc0 calls this function
        # whereas worker procs call Dofs.f().
        f_unshifted = self.dofs.f()
        return self.f_from_unshifted(f_unshifted)

    def objective_from_shifted_f(self, f):
        """
        Given a vector of functions that has already been evaluated,
        and already shifted and scaled, convert the result to the
        overall scalar objective function, without any further
        function evaluations. This routine is useful if we have
        already evaluated the residuals and we want to convert the
        result to the overall scalar objective without the
        computational expense of further function evaluations.
        """
        assert len(f) == self.dofs.nvals
        return np.dot(f, f)

    def objective_from_unshifted_f(self, f_unshifted):
        """
        Given a vector of functions that has already been evaluated,
        but not yet shifted and scaled, convert the result to the
        overall scalar objective function, without any further
        function evaluations. This routine is useful if we have
        already evaluated the residuals and we want to convert the
        result to the overall scalar objective without the
        computational expense of further function evaluations.
        """
        assert len(f_unshifted) == self.dofs.nvals
        f_shifted = self.f_from_unshifted(f_unshifted)
        return self.objective_from_shifted_f(f_shifted)

    def scale_dofs_jac(self, jmat):
        """
        Given a Jacobian matrix j for the Dofs() associated to this
        least-squares problem, return the scaled Jacobian matrix for
        the least-squares residual terms. This function does not
        actually compute the Dofs() Jacobian, since sometimes we would
        compute that directly whereas other times we might compute it
        with serial or parallel finite differences. The provided jmat
        is scaled in-place.
        """
        logger.info("scale_dofs_jac() called")

        # Scale rows by sqrt(weight):
        start_index = 0
        for j in range(self.dofs.nfuncs):
            end_index = start_index + self.dofs.nvals_per_func[j]
            #jac[j, :] = jac[j, :] * np.sqrt(self.terms[j].weight)
            jmat[start_index:end_index, :] = jmat[start_index:end_index, :] \
                * np.sqrt(self.terms[j].weight)
            start_index = end_index

        return np.array(jmat)

    def jac(self, x=None, **kwargs):
        """
        This method gives the Jacobian of the residuals with respect to
        the parameters, if it is available, given the state vector
        x. This function is passed to scipy.optimize, and could be
        passed to other optimization algorithms too. This Jacobian
        differs from the one returned by Dofs() because it accounts
        for the 'weight' scale factors.

        If the argument x is not supplied, the Jacobian will be
        evaluated for the present state vector. If x is supplied, then
        first set_dofs() will be called for each object to set the
        global state vector to x.

        kwargs is passed to Dofs.fd_jac().
        """
        logger.info("jac() called with x=" + str(x))

        self.x = x

        # This next bit does the hard work of evaluating the
        # Jacobian.
        # Bharat's comment: The conditional logic should be delegated to
        # Bharat's comment: Dofs class
        if self.dofs.grad_avail:
            logger.debug('Calling analytic Jacobian')
            jmat = self.dofs.jac()
        else:
            logger.debug('Calling finite_difference Jacobian')
            jmat = self.dofs.fd_jac(**kwargs)

        # Scale by sqrt(weight) factor:
        return self.scale_dofs_jac(jmat)



import numpy as np
import simsoptpp as sopp
from .surface import Surface


class SurfaceRZFourier(sopp.SurfaceRZFourier, Surface):
    r"""`SurfaceRZFourier` is a surface that is represented in cylindrical
    coordinates using the following Fourier series:

    .. math::
           r(\theta, \phi) = \sum_{m=0}^{m_{\text{pol}}} \sum_{n=-n_{\text{tor}}}^{n_\text{tor}} [
               r_{c,m,n} \cos(m \theta - n_{\text{fp}} n \phi)
               + r_{s,m,n} \sin(m \theta - n_{\text{fp}} n \phi) ]

    and the same for :math:`z(\theta, \phi)`.

    Here, :math:`(r,\phi, z)` are standard cylindrical coordinates, and theta
    is any poloidal angle.

    Note that for :math:`m=0` we skip the :math:`n<0` term for the cos terms, and the :math:`n \leq 0`
    for the sin terms.

    In addition, in the ``stellsym=True`` case, we skip the sin terms for :math:`r`, and
    the cos terms for :math:`z`.
    """

    def __init__(self, nfp=1, stellsym=True, mpol=1, ntor=0, quadpoints_phi=63, quadpoints_theta=62):
        if isinstance(quadpoints_phi, np.ndarray):
            quadpoints_phi = list(quadpoints_phi)
            quadpoints_theta = list(quadpoints_theta)
        sopp.SurfaceRZFourier.__init__(self, mpol, ntor, nfp, stellsym, quadpoints_phi, quadpoints_theta)
        self.rc[0, ntor] = 1.0
        self.rc[1, ntor] = 0.1
        self.zs[1, ntor] = 0.1
        Surface.__init__(self)
        self.make_names()

    def get_dofs(self):
        """
        Return the dofs associated to this surface.
        """
        return np.asarray(sopp.SurfaceRZFourier.get_dofs(self))

    def make_names(self):
        """
        Form a list of names of the `rc`, `zs`, `rs`, or `zc` array elements.
        """
        self.names = self.make_names_helper('rc', True) + self.make_names_helper('zs', False)
        if not self.stellsym:
            self.names += self.make_names_helper('rs', False) + self.make_names_helper('zc', True)

    def make_names_helper(self, prefix, include0):
        if include0:
            names = [prefix + "(0,0)"]
        else:
            names = []

        names += [prefix + '(0,' + str(n) + ')' for n in range(1, self.ntor + 1)]
        for m in range(1, self.mpol + 1):
            names += [prefix + '(' + str(m) + ',' + str(n) + ')' for n in range(-self.ntor, self.ntor + 1)]
        return names

    @classmethod
    def from_focus(cls, filename, nphi=32, ntheta=32):
        """
        Read in a surface from a FOCUS-format file.
        """
        f = open(filename, 'r')
        lines = f.readlines()
        f.close()

        # Read the line containing Nfou and nfp:
        splitline = lines[1].split()
        errmsg = "This does not appear to be a FOCUS-format file."
        assert len(splitline) == 3, errmsg
        Nfou = int(splitline[0])
        nfp = int(splitline[1])

        # Now read the Fourier amplitudes:
        n = np.full(Nfou, 0)
        m = np.full(Nfou, 0)
        rc = np.zeros(Nfou)
        rs = np.zeros(Nfou)
        zc = np.zeros(Nfou)
        zs = np.zeros(Nfou)
        for j in range(Nfou):
            splitline = lines[j + 4].split()
            n[j] = int(splitline[0])
            m[j] = int(splitline[1])
            rc[j] = float(splitline[2])
            rs[j] = float(splitline[3])
            zc[j] = float(splitline[4])
            zs[j] = float(splitline[5])
        assert np.min(m) == 0
        stellsym = np.max(np.abs(rs)) == 0 and np.max(np.abs(zc)) == 0
        mpol = int(np.max(m))
        ntor = int(np.max(np.abs(n)))

        surf = cls(mpol=mpol, ntor=ntor, nfp=nfp, stellsym=stellsym, quadpoints_phi=nphi, quadpoints_theta=ntheta)
        for j in range(Nfou):
            surf.rc[m[j], n[j] + ntor] = rc[j]
            surf.zs[m[j], n[j] + ntor] = zs[j]
            if not stellsym:
                surf.rs[m[j], n[j] + ntor] = rs[j]
                surf.zc[m[j], n[j] + ntor] = zc[j]

        return surf

    def change_resolution(self, mpol, ntor):
        """
        Change the values of `mpol` and `ntor`. Any new Fourier amplitudes
        will have a magnitude of zero.  Any previous nonzero Fourier
        amplitudes that are not within the new range will be
        discarded.
        """
        old_mpol = self.mpol
        old_ntor = self.ntor
        old_rc = self.rc
        old_zs = self.zs
        if not self.stellsym:
            old_rs = self.rs
            old_zc = self.zc
        self.mpol = mpol
        self.ntor = ntor
        self.allocate()
        if mpol < old_mpol or ntor < old_ntor:
            self.invalidate_cache()

        min_mpol = np.min((mpol, old_mpol))
        min_ntor = np.min((ntor, old_ntor))
        for m in range(min_mpol + 1):
            for n in range(-min_ntor, min_ntor + 1):
                self.rc[m, n + ntor] = old_rc[m, n + old_ntor]
                self.zs[m, n + ntor] = old_zs[m, n + old_ntor]
                if not self.stellsym:
                    self.rs[m, n + ntor] = old_rs[m, n + old_ntor]
                    self.zc[m, n + ntor] = old_zc[m, n + old_ntor]
        self.make_names()

    def to_RZFourier(self):
        """
        No conversion necessary.
        """
        return self

    def __repr__(self):
        return "SurfaceRZFourier " + str(hex(id(self))) + " (nfp=" + \
            str(self.nfp) + ", stellsym=" + str(self.stellsym) + \
            ", mpol=" + str(self.mpol) + ", ntor=" + str(self.ntor) \
            + ")"

    def _validate_mn(self, m, n):
        """
        Check whether `m` and `n` are in the allowed range.
        """
        if m < 0:
            raise ValueError('m must be >= 0')
        if m > self.mpol:
            raise ValueError('m must be <= mpol')
        if n > self.ntor:
            raise ValueError('n must be <= ntor')
        if n < -self.ntor:
            raise ValueError('n must be >= -ntor')

    def get_rc(self, m, n):
        """
        Return a particular `rc` Parameter.
        """
        self._validate_mn(m, n)
        return self.rc[m, n + self.ntor]

    def get_rs(self, m, n):
        """
        Return a particular `rs` Parameter.
        """
        if self.stellsym:
            return ValueError(
                'rs does not exist for this stellarator-symmetric surface.')
        self._validate_mn(m, n)
        return self.rs[m, n + self.ntor]

    def get_zc(self, m, n):
        """
        Return a particular `zc` Parameter.
        """
        if self.stellsym:
            return ValueError(
                'zc does not exist for this stellarator-symmetric surface.')
        self._validate_mn(m, n)
        return self.zc[m, n + self.ntor]

    def get_zs(self, m, n):
        """
        Return a particular `zs` Parameter.
        """
        self._validate_mn(m, n)
        return self.zs[m, n + self.ntor]

    def set_rc(self, m, n, val):
        """
        Set a particular `rc` Parameter.
        """
        self._validate_mn(m, n)
        self.rc[m, n + self.ntor] = val
        self.recalculate = True
        self.recalculate_derivs = True

    def set_rs(self, m, n, val):
        """
        Set a particular `rs` Parameter.
        """
        if self.stellsym:
            return ValueError(
                'rs does not exist for this stellarator-symmetric surface.')
        self._validate_mn(m, n)
        self.rs[m, n + self.ntor] = val
        self.recalculate = True
        self.recalculate_derivs = True

    def set_zc(self, m, n, val):
        """
        Set a particular `zc` Parameter.
        """
        if self.stellsym:
            return ValueError(
                'zc does not exist for this stellarator-symmetric surface.')
        self._validate_mn(m, n)
        self.zc[m, n + self.ntor] = val
        self.recalculate = True
        self.recalculate_derivs = True

    def set_zs(self, m, n, val):
        """
        Set a particular `zs` Parameter.
        """
        self._validate_mn(m, n)
        self.zs[m, n + self.ntor] = val
        self.recalculate = True
        self.recalculate_derivs = True

    def fixed_range(self, mmin, mmax, nmin, nmax, fixed=True):
        """
        Set the 'fixed' property for a range of `m` and `n` values.

        All modes with `m` in the interval [`mmin`, `mmax`] and `n` in the
        interval [`nmin`, `nmax`] will have their fixed property set to
        the value of the `fixed` parameter. Note that `mmax` and `nmax`
        are included (unlike the upper bound in python's range(min,
        max).)
        """
        for m in range(mmin, mmax + 1):
            this_nmin = nmin
            if m == 0 and nmin < 0:
                this_nmin = 0
            for n in range(this_nmin, nmax + 1):
                self.set_fixed('rc({},{})'.format(m, n), fixed)
                if m > 0 or n != 0:
                    self.set_fixed('zs({},{})'.format(m, n), fixed)
                if not self.stellsym:
                    self.set_fixed('zc({},{})'.format(m, n), fixed)
                    if m > 0 or n != 0:
                        self.set_fixed('rs({},{})'.format(m, n), fixed)

    def to_Garabedian(self):
        """
        Return a `SurfaceGarabedian` object with the identical shape.

        For a derivation of the transformation here, see
        https://terpconnect.umd.edu/~mattland/assets/notes/toroidal_surface_parameterizations.pdf
        """
        if not self.stellsym:
            raise RuntimeError('Non-stellarator-symmetric SurfaceGarabedian objects have not been implemented')
        from simsopt.geo.surfacegarabedian import SurfaceGarabedian
        mmax = self.mpol + 1
        mmin = np.min((0, 1 - self.mpol))
        s = SurfaceGarabedian(nfp=self.nfp, mmin=mmin, mmax=mmax, nmin=-self.ntor, nmax=self.ntor)
        for n in range(-self.ntor, self.ntor + 1):
            for m in range(mmin, mmax + 1):
                Delta = 0
                if m - 1 >= 0:
                    Delta = 0.5 * (self.get_rc(m - 1, n) - self.get_zs(m - 1, n))
                if 1 - m >= 0:
                    Delta += 0.5 * (self.get_rc(1 - m, -n) + self.get_zs(1 - m, -n))
                s.set_Delta(m, n, Delta)

        return s

    def set_dofs(self, dofs):
        sopp.SurfaceRZFourier.set_dofs(self, dofs)
        for d in self.dependencies:
            d.invalidate_cache()

    def darea(self):
        """
        Short hand for `Surface.darea_by_dcoeff()`
        """
        return self.darea_by_dcoeff()

    def dvolume(self):
        """
        Short hand for `Surface.dvolume_by_dcoeff()`
        """
        return self.dvolume_by_dcoeff()



# coding: utf-8
# Copyright (c) HiddenSymmetries Development Team.
# Distributed under the terms of the LGPL License

"""
This module provides a class that handles the transformation to
Boozer coordinates, and an optimization target for quasisymmetry.
"""

import logging
from typing import Union, Iterable

import numpy as np
from ..util.dev import SimsoptRequires

logger = logging.getLogger(__name__)

try:
    from mpi4py import MPI
except ImportError as e:
    MPI = None 
    logger.warning(str(e))

# booz_xform_found = True
try:
    import booz_xform
except ImportError as e:
    # booz_xform_found = False
    booz_xform = None
    logger.warning(str(e))

from .._core.optimizable import Optimizable

#try:
if MPI is not None:
    from .vmec import Vmec
#except ImportError as e:
    #Vmec = None
    #logger.warning(str(e))


#@requires(booz_xform is not None,
#          "To use a Boozer object, the booz_xform package "
#          "must be installed. Run 'pip install -v booz_xform'")
@SimsoptRequires(MPI is not None, "mpi4py needs to be installed for running booz-xform")
class Boozer(Optimizable):
    """
    This class handles the transformation to Boozer coordinates.

    A Boozer instance maintains a set "s", which is a registry of the
    surfaces on which other objects want Boozer-coordinate data. When
    the run() method is called, the Boozer transformation is carried
    out on all these surfaces. The registry can be cleared at any time
    by setting the s attribute to {}.
    """

    def __init__(self,
                 equil: Vmec,
                 mpol: int = 32,
                 ntor: int = 32) -> None:
        """
        Constructor
        """
        if booz_xform is None:
            raise RuntimeError(
                "To use a Boozer object, the booz_xform package "
                "must be installed. Run 'pip install -v booz_xform'")

        self.equil = equil
        self.depends_on = ["equil"]
        self.mpol = mpol
        self.ntor = ntor
        self.bx = booz_xform.Booz_xform()
        self.s = set()
        self.need_to_run_code = True
        self._calls = 0  # For testing, keep track of how many times we call bx.run()

        # We may at some point want to allow booz_xform to use a
        # different partitioning of the MPI processors compared to the
        # equilibrium code. But for simplicity, we'll use the same mpi
        # partition for now. For unit tests, we allow equil to be None,
        # so we have to allow for this case here.
        self.mpi = None
        if equil is not None:
            self.mpi = equil.mpi

    def get_dofs(self):
        return np.array([])

    def set_dofs(self, x):
        self.need_to_run_code = True

    def register(self, s: Union[float, Iterable[float]]) -> None:
        """
        This function is called by objects that depend on this Boozer
        object, to indicate that they will want Boozer data on the
        given set of surfaces.

        Args:
            s: 1 or more surfaces on which Boozer data will be requested.
        """
        # Force input surface data to be a set:
        try:
            ss = set(s)
        except:
            ss = {s}

        for new_s in ss:
            if new_s < 0 or new_s > 1:
                raise ValueError("Normalized toroidal flux values s must lie"
                                 "in the interval [0, 1]")
        logger.info("Adding entries to Boozer registry: {}".format(ss))
        self.s = self.s.union(ss)
        self.need_to_run_code = True

    def run(self):
        """
        Run booz_xform on all the surfaces that have been registered.
        """

        if (self.mpi is not None) and (not self.mpi.proc0_groups):
            logger.info("This proc is skipping Boozer.run since it is not a group leader.")
            return

        if not self.need_to_run_code:
            logger.info("Boozer.run() called but no need to re-run Boozer transformation.")
            return

        s = sorted(list(self.s))
        logger.info("Preparing to run Boozer transformation. Registry:{}".format(s))

        if isinstance(self.equil, Vmec):
            self.equil.run()
            wout = self.equil.wout  # Shorthand

            # Get the half-grid points that are closest to the requested values
            ns = wout.ns
            s_full = np.linspace(0, 1, ns)
            ds = s_full[1] - s_full[0]
            s_half = s_full[1:] - 0.5 * ds

            # For each float value of s at which the Boozer results
            # have been requested, we need to find the corresponding
            # radial index of the booz_xform results. The result is
            # self.s_to_index. Computing this is tricky because
            # multiple values of s may get rounded to the same
            # half-grid surface. The solution here is done in two
            # steps. First we find a map from each float value of s to
            # the corresponding radial index among all half-grid
            # surfaces (even ones where we won't compute the Boozer
            # transformation.) This resulting map is
            # s_to_index_all_surfs. In a second step,
            # s_to_index_all_surfs and the list of compute_surfs are
            # used to find s_to_index.

            compute_surfs = []
            s_to_index_all_surfs = dict()
            self.s_used = dict()
            for ss in s:
                index = np.argmin(np.abs(s_half - ss))
                compute_surfs.append(index)
                s_to_index_all_surfs[ss] = index
                self.s_used[ss] = s_half[index]

            # Eliminate any duplicates
            compute_surfs = sorted(list(set(compute_surfs)))
            logger.info("compute_surfs={}".format(compute_surfs))
            logger.info("s_to_index_all_surfs={}".format(s_to_index_all_surfs))
            self.s_to_index = dict()
            for ss in s:
                self.s_to_index[ss] = compute_surfs.index(s_to_index_all_surfs[ss])
            logger.info("s_to_index={}".format(self.s_to_index))

            # Transfer data in memory from VMEC to booz_xform
            self.bx.asym = bool(wout.lasym)
            self.bx.nfp = wout.nfp

            self.bx.mpol = wout.mpol
            self.bx.ntor = wout.ntor
            self.bx.mnmax = wout.mnmax
            self.bx.xm = wout.xm
            self.bx.xn = wout.xn
            print('mnmax:', wout.mnmax, ' len(xm):', len(wout.xm), ' len(xn):', len(wout.xn))
            print('mnmax_nyq:', wout.mnmax_nyq, ' len(xm_nyq):', len(wout.xm_nyq), ' len(xn_nyq):', len(wout.xn_nyq))
            assert len(wout.xm) == wout.mnmax
            assert len(wout.xn) == wout.mnmax
            assert len(self.bx.xm) == self.bx.mnmax
            assert len(self.bx.xn) == self.bx.mnmax

            self.bx.mpol_nyq = int(wout.xm_nyq[-1])
            self.bx.ntor_nyq = int(wout.xn_nyq[-1] / wout.nfp)
            self.bx.mnmax_nyq = wout.mnmax_nyq
            self.bx.xm_nyq = wout.xm_nyq
            self.bx.xn_nyq = wout.xn_nyq
            assert len(wout.xm_nyq) == wout.mnmax_nyq
            assert len(wout.xn_nyq) == wout.mnmax_nyq
            assert len(self.bx.xm_nyq) == self.bx.mnmax_nyq
            assert len(self.bx.xn_nyq) == self.bx.mnmax_nyq

            if wout.lasym:
                rmns = wout.rmns
                zmnc = wout.zmnc
                lmnc = wout.lmnc
                bmns = wout.bmns
                bsubumns = wout.bsubumns
                bsubvmns = wout.bsubvmns
            else:
                # For stellarator-symmetric configs, the asymmetric
                # arrays have not been initialized.
                arr = np.array([[]])
                rmns = arr
                zmnc = arr
                lmnc = arr
                bmns = arr
                bsubumns = arr
                bsubvmns = arr

            # For quantities that depend on radius, booz_xform handles
            # interpolation and discarding the rows of zeros:
            self.bx.init_from_vmec(wout.ns,
                                   wout.iotas,
                                   wout.rmnc,
                                   rmns,
                                   zmnc,
                                   wout.zmns,
                                   lmnc,
                                   wout.lmns,
                                   wout.bmnc,
                                   bmns,
                                   wout.bsubumnc,
                                   bsubumns,
                                   wout.bsubvmnc,
                                   bsubvmns)
            self.bx.compute_surfs = compute_surfs
            self.bx.mboz = self.mpol
            self.bx.nboz = self.ntor            

        else:
            # Cases for SPEC, GVEC, etc could be added here.
            raise ValueError("equil is not an equilibrium type supported by"
                             "Boozer")

        logger.info("About to call booz_xform.Booz_xform.run().")
        self.bx.run()
        self._calls += 1
        logger.info("Returned from calling booz_xform.Booz_xform.run().")
        self.need_to_run_code = False


class Quasisymmetry(Optimizable):
    """
    This class is used to compute the departure from quasisymmetry on
    a given flux surface based on the Boozer spectrum.

    Args:
        boozer: A Boozer object on which the calculation will be based.
        s: The normalized toroidal magnetic flux for the flux surface to analyze. Should be in the range [0, 1].
        m: The poloidal mode number of the symmetry you want to achive.
           The departure from symmetry B(m * theta - nfp * n * zeta) will be reported.
        n: The toroidal mode number of the symmetry you want to achieve.
           The departure from symmetry B(m * theta - nfp * n * zeta) will be reported.
        normalization: A uniform normalization applied to all bmnc harmonics.
           If ``"B00"``, the symmetry-breaking modes will be divided by the m=n=0 mode amplitude
           on the same surface. If ``"symmetric"``, the symmetry-breaking modes will be
           divided by the square root of the sum of the squares of all the symmetric
           modes on the same surface. This is the normalization used in stellopt.
        weight: An option for a m- or n-dependent weight to be applied to the bmnc amplitudes.
    """

    def __init__(self,
                 boozer: Boozer,
                 s: Union[float, Iterable[float]],
                 m: int,
                 n: int,
                 normalization: str = "B00",
                 weight: str = "even") -> None:
        """
        Constructor

        """
        self.boozer = boozer
        self.m = m
        self.n = n
        self.normalization = normalization
        self.weight = weight
        self.depends_on = ['boozer']

        # If s is not already iterable, make it so:
        try:
            iter(s)
        except:
            s = [s]
        self.s = s
        boozer.register(s)

    def get_dofs(self):
        return np.array([])

    def set_dofs(self, x):
        self.need_to_run_code = True

    def J(self) -> np.ndarray:
        """
        Carry out the calculation of the quasisymmetry error.

        Returns:
            1D numpy array listing all the normalized mode amplitudes of 
            symmetry-breaking Fourier modes of ``|B|``.
        """

        # Only group leaders do anything:
        if (self.boozer.mpi is not None) and (not self.boozer.mpi.proc0_groups):
            logger.info("This proc is skipping Quasisymmetry.J since it is not a group leader.")
            return np.array([])

        # The next line is the expensive part of the calculation:
        self.boozer.run()

        symmetry_error = []
        for js, s in enumerate(self.s):
            index = self.boozer.s_to_index[s]
            bmnc = self.boozer.bx.bmnc_b[:, index]
            xm = self.boozer.bx.xm_b
            xn = self.boozer.bx.xn_b / self.boozer.bx.nfp

            if self.m != 0 and self.m != 1:
                raise ValueError("m for quasisymmetry should be 0 or 1.")

            # Find the indices of the symmetric modes:
            if self.n == 0:
                # Quasi-axisymmetry
                symmetric = (xn == 0)

            elif self.m == 0:
                # Quasi-poloidal symmetry
                symmetric = (xm == 0)

            else:
                # Quasi-helical symmetry
                symmetric = (xm * self.n + xn * self.m == 0)
                # Stellopt takes the "and" of this with mod(xm, self.m),
                # which does not seem necessary since self.m must be 1 to
                # get here.
            nonsymmetric = np.logical_not(symmetric)

            # Scale all bmnc modes so the average |B| is 1 or close to 1:

            if self.normalization == "B00":
                # Normalize by the (m,n) = (0,0) mode amplitude:
                assert xm[0] == 0
                assert xn[0] == 0
                bnorm = bmnc[0]

            elif self.normalization == "symmetric":
                # Normalize by sqrt(sum_{symmetric modes} B{m,n}^2)
                temp = bmnc[symmetric]
                bnorm = np.sqrt(np.dot(temp, temp))

            else:
                raise ValueError("Unrecognized value for normalization in Quasisymmetry")

            logger.info("For s={}, bnorm={}".format(s, bnorm))
            bmnc = bmnc / bnorm

            # Apply any weight that depends on m and/or n:

            if self.weight == "even":
                # Evenly weight each bmnc mode. Normalize by the m=n=0 mode on that surface.
                symmetry_error.append(bmnc[nonsymmetric])

            elif self.weight == "stellopt":
                # Stellopt appears to apply a m-dependent radial
                # weight, assuming sigma > 0. However, the m is
                # evaluated outside of any loop over m, so m ends up
                # taking the value mboz instead of the actual m for
                # each mode. As a result, there is an even weight by s_used**2.

                s_used = self.boozer.s_used[s]
                logger.info('s_used, in Quasisymmetry: {}'.format(s_used))
                """
                rad_sigma = np.full(len(xm), s_used * s_used)
                rad_sigma[xm < 3] = s_used
                rad_sigma[xm == 3] = s_used ** 1.5
                """
                rad_sigma = s_used * s_used
                temp = bmnc / rad_sigma
                symmetry_error.append(temp[nonsymmetric])

            elif self.weight == "stellopt_ornl":
                # This option is similar to "stellopt" except we
                # return a single number for the residual instead of a
                # vector of residuals.

                # For this option, stellopt applies a m-dependent
                # radial weight only when sigma < 0, the opposite of
                # when using the non-ORNL helicity! Here, we do not
                # apply such a weight.

                temp = bmnc[nonsymmetric]
                symmetry_error.append(np.array([np.sqrt(np.sum(temp * temp))]))

            else:
                raise ValueError("Unrecognized value for weight in Quasisymmetry")

        return np.array(symmetry_error).flatten()
